import os
import logging
import numpy as np
from tqdm.auto import tqdm
from collections import OrderedDict
from batchgenerators.dataloading import MultiThreadedAugmenter
from .callbacks import AbstractCallback
from .abstract_trainer import AbstractNetworkTrainer

from delira import get_backends

logger = logging.getLogger(__name__)

if "TORCH" in get_backends():
    import torch
    from .train_utils import pytorch_batch_to_numpy
    from .train_utils import create_optims_default_pytorch as create_optims_default
    from ..io.torch import load_checkpoint, save_checkpoint

    class PyTorchNetworkTrainer(AbstractNetworkTrainer):
        """
        Train and Validate a Network

        See Also
        --------
        :class:`AbstractNetwork`

        """

        def __init__(self, network, save_path,
                     criterions: dict, optimizer_cls,
                     optimizer_params={}, metrics={}, lr_scheduler_cls=None,
                     lr_scheduler_params={}, gpu_ids=[], save_freq=1,
                     optim_fn=create_optims_default,
                     fold=0, callbacks=[], start_epoch=1, mixed_precision=False,
                     mixed_precision_kwargs={"enable_caching": True,
                                             "verbose": False,
                                             "allow_banned": False},
                     **kwargs):
            """

            Parameters
            ----------
            network : :class:`AbstractPyTorchNetwork`
                the network to train
            save_path : str
                path to save networks to
            criterions : dict
                dictionary containing the training criterions
            optimizer_cls : subclass of torch.optim.Optimizer
                optimizer class implementing the optimization algorithm of choice
            optimizer_params : dict
                keyword arguments passed to optimizer during construction
            metrics : dict
                dictionary containing the validation metrics
            lr_scheduler_cls : Any
                learning rate schedule class: must implement step() method
            lr_scheduler_params : dict
                keyword arguments passed to lr scheduler during construction
            gpu_ids : list
                list containing ids of GPUs to use; if empty: use cpu instead
            save_freq : int
                integer specifying how often to save the current model's state.
                State is saved every state_freq epochs
            optim_fn : function
                creates a dictionary containing all necessary optimizers
            fold : int
                current cross validation fold (0 per default)
            callbacks : list
                initial callbacks to register
            start_epoch : int
                epoch to start training at
            mixed_precision : bool
                whether to use mixed precision or not (False per default)
            mixed_precision_kwargs : dict
                additional keyword arguments for mixed precision
            **kwargs :
                additional keyword arguments

            """

            super().__init__(fold, callbacks)

            self.save_path = save_path
            if os.path.isdir(save_path):
                logger.warning(
                    "Save Path already exists. Saved Models may be overwritten")
            else:
                os.makedirs(save_path)

            self.criterions = criterions

            self.metrics = metrics

            self.save_freq = save_freq

            # Whether or not to stop the training
            # Used for early stopping
            self.stop_training = False
            self.start_epoch = start_epoch

            self._setup(network, optim_fn, optimizer_cls, optimizer_params,
                        lr_scheduler_cls, lr_scheduler_params, gpu_ids,
                        mixed_precision, mixed_precision_kwargs)

            for key, val in kwargs.items():
                setattr(self, key, val)

        def _setup(self, network, optim_fn, optimizer_cls, optimizer_params,
                   lr_scheduler_cls, lr_scheduler_params, gpu_ids,
                   mixed_precision, mixed_precision_kwargs):
            """
            Defines the Trainers Setup

            Parameters
            ----------
            network : :class:`AbstractPyTorchNetwork`
                the network to train
            optim_fn : function
                creates a dictionary containing all necessary optimizers
            optimizer_cls : subclass of torch.optim.Optimizer
                optimizer class implementing the optimization algorithm of choice
            optimizer_params : dict
            lr_scheduler_cls : Any
                learning rate schedule class: must implement step() method
            lr_scheduler_params : dict
                keyword arguments passed to lr scheduler during construction
            gpu_ids : list
                list containing ids of GPUs to use; if empty: use cpu instead
            mixed_precision : bool
                whether to use mixed precision or not (False per default)
            mixed_precision_kwargs : dict
                additional keyword arguments for mixed precision

            """

            try:
                from apex import amp
                self._amp_handle = amp.init(mixed_precision,
                                            *mixed_precision_kwargs)
                wrap_fn = self._amp_handle.wrap_optimizer

            except ImportError:
                if mixed_precision:
                    logger.warning("Apex was not found found, trying to continue \
                                    in full precision instead")
                from ..utils.context_managers import DefaultOptimWrapperTorch
                wrap_fn = DefaultOptimWrapperTorch

            # wrap optimizers by half_precision_optimizer via apex if necessary
            self.optimizers = {k: wrap_fn(
                v, num_loss=len(self.criterions)) for k, v
                in optim_fn(network, optimizer_cls, **optimizer_params).items()}

            # schedulers
            if lr_scheduler_cls is not None:
                for key, optim in self.optimizers.items():
                    if not issubclass(lr_scheduler_cls, AbstractCallback):
                        logger.warning("lr_scheduler_cls is not a callback.")
                    # access actual optimizer by calling wrapped optimizer from wrapper
                    self.register_callback(lr_scheduler_cls(optim._optimizer,
                                                            **lr_scheduler_params))
                    
            # store network in self.module to load previous state
            # (will be overwritten later)
            self.module = network


            # Load latest epoch file if available
            if os.path.isdir(self.save_path):
                # check all files in directory starting with "checkpoint" and not
                # ending with "_best.pth"
                files = [x for x in os.listdir(self.save_path)
                         if os.path.isfile(os.path.join(self.save_path, x))
                         and x.startswith("checkpoint")
                         and not x.endswith("_best.pth")]

                # if list is not empty: load previous state
                if files:

                    latest_epoch = max([int(x.rsplit("_", 1)[-1].rsplit(".", 1)[0])
                                        for x in files])

                    latest_state_path = os.path.join(self.save_path,
                                                     "checkpoint_epoch_%d.pth"
                                                     % latest_epoch)

                    logger.info("Attempting to load state from previous \
                                training from %s" % latest_state_path)
                    try:
                        self.update_state(latest_state_path)
                    except KeyError:
                        logger.warn("Previous State could not be loaded, \
                                    although it exists.Training will be \
                                    restarted")

            # asssign closure and prepare batch from network
            self.closure_fn = network.closure
            self._prepare_batch = network.prepare_batch

            if gpu_ids and torch.cuda.is_available():
                self.use_gpu = True
                if (len(gpu_ids) > 1) and (torch.cuda.device_count() > 1):
                    # use GPU 0 as default input GPU
                    self.input_device = torch.device("cuda:%d" % gpu_ids[0])

                    # Train on multiple GPUs and use GPU 0 as output device
                    self.module = torch.nn.DataParallel(self.module.to(
                        self.input_device),
                        device_ids=gpu_ids,
                        output_device=gpu_ids[1])

                    # use GPU 1 as default output GPU for balanced GPU usage
                    self.output_device = torch.device("cuda:%d" % gpu_ids[1])
                else:
                    # use the only available GPU as input device
                    self.input_device = torch.device("cuda:%d" % gpu_ids[0])
                    self.module = self.module.to(self.input_device)

                    # use GPU 0 as output device as output device
                    self.output_device = torch.device("cuda:%d" % gpu_ids[0])
            else:
                self.use_gpu = False
                self.input_device = torch.device("cpu")
                self.output_device = torch.device("cpu")
                self.module = self.module.to(self.input_device)

        def train(self, num_epochs, datamgr_train, datamgr_valid=None,
                  val_score_key=None, val_score_mode='highest'):
            """
            train network

            Parameters
            ----------
            num_epochs : int
                number of epochs to train
            datamgr_train : BaseDataManager
                Data Manager to create Batch Generator for training
            datamgr_valid : BaseDataManager
                Data Manager to create Batch Generator for validation
            val_score_key : str
                Key of validation metric; must be key in self.metrics
            val_score_mode : str
                String to specify whether a higher or lower validation score is
                optimal; must be in ['highest', 'lowest']

            Returns
            -------
            :class:`AbstractPyTorchNetwork`
                Best model (if `val_score_key` is not a valid key the model of the
                last epoch will be returned)

            """

            self._at_training_begin()

            self.module.train()

            if val_score_mode == 'highest':
                best_val_score = 0
            elif val_score_mode == 'lowest':
                best_val_score = float('inf')
            else:
                best_val_score = None

            curr_val_score = best_val_score

            self.save_state(os.path.join(self.save_path, "checkpoint_epoch_0.pth"),
                            self.start_epoch)
            metrics_val = {}

            for epoch in range(self.start_epoch, num_epochs+1):

                self._at_epoch_begin(metrics_val, val_score_key, epoch,
                                     num_epochs)

                batch_gen_train = datamgr_train.get_batchgen(seed=epoch)

                self._train_single_epoch(batch_gen_train, epoch)

                if datamgr_valid:
                    # validate with batchsize 1 and 1 augmentation processs to
                    # avoid dropping of last elements
                    orig_num_aug_processes = datamgr_valid.n_process_augmentation
                    orig_batch_size = datamgr_valid.batch_size

                    datamgr_valid.batch_size = 1
                    datamgr_valid.n_process_augmentation = 1

                    pred_val, labels_val, metrics_val = self.predict(
                        datamgr_valid.get_batchgen(), batch_size=orig_batch_size)

                    # reset old values
                    datamgr_valid.batch_size = orig_batch_size
                    datamgr_valid.n_process_augmentation = orig_num_aug_processes

                    # ToDO: Move decision, if current model is best to callback
                    if val_score_key in metrics_val.keys():
                        curr_val_score = metrics_val[val_score_key]
                        is_best = self._is_better_val_scores(best_val_score,
                                                             curr_val_score,
                                                             val_score_mode)

                    else:
                        logger.warning(
                            "Validation score key not in metric dict. "
                            "Logging metrics but can't decide which model is best")

                        is_best = False

                    if is_best:
                        best_val_score = curr_val_score
                        tqdm.write(
                            'Best val score = %2.3f' % best_val_score.item())
                    else:
                        is_best = False
                else:
                    is_best = False
                    labels_val, pred_val, metrics_val = {}, {}, {}

                self._at_epoch_end(metrics_val, val_score_key, epoch, is_best)

                # stop training (might be caused by early stopping)
                if self.stop_training:
                    break

            return self._at_training_end()

        def _at_training_begin(self, *args, **kwargs):
            """
            Defines behaviour at beginning of training

            Parameters
            ----------
            *args :
                positional arguments
            **kwargs :
                keyword arguments

            """
            pass

        def _at_training_end(self):
            """
            Defines Behaviour at end of training: Loads best model if available

            Returns
            -------
            :class:`AbstractPyTorchNetwork`
                best network

            """
            if os.path.isfile(os.path.join(self.save_path, 'checkpoint_best.pth')):

                # load best model and return it
                self.update_state(os.path.join(self.save_path,
                                               'checkpoint_best.pth')
                                  )

            return self.module

        def _at_epoch_begin(self, metrics_val, val_score_key, epoch, num_epochs,
                            **kwargs):
            """
            Defines behaviour at beginning of each epoch: Executes all callbacks's
            `at_epoch_begin` method

            Parameters
            ----------
            metrics_val : dict
                validation metrics
            val_score_key : str
                validation score key
            epoch : int
                current epoch
            num_epochs : int
                total number of epochs
            **kwargs :
                keyword arguments

            """

            # execute all callbacks
            for cb in self._callbacks:
                self._update_state(cb.at_epoch_begin(self, val_metrics=metrics_val,
                                                     val_score_key=val_score_key,
                                                     curr_epoch=epoch))

        def _at_epoch_end(self, metrics_val, val_score_key, epoch, is_best,
                          **kwargs):
            """
            Defines behaviour at beginning of each epoch: Executes all callbacks's
            `at_epoch_end` method and saves current state if necessary

            Parameters
            ----------
            metrics_val : dict
                validation metrics
            val_score_key : str
                validation score key
            epoch : int
                current epoch
            num_epochs : int
                total number of epochs
            is_best : bool
                whether current model is best one so far
            **kwargs :
                keyword arguments

            """

            for cb in self._callbacks:
                self._update_state(cb.at_epoch_end(self, val_metrics=metrics_val,
                                                   val_score_key=val_score_key,
                                                   curr_epoch=epoch))

            if epoch % self.save_freq == 0:
                self.save_state(os.path.join(self.save_path,
                                             "checkpoint_epoch_%d.pth" % epoch),
                                epoch)

            if is_best:
                self.save_state(os.path.join(self.save_path,
                                             "checkpoint_best.pth"),
                                epoch)

        def _train_single_epoch(self, batchgen: MultiThreadedAugmenter, epoch):
            """
            Trains the network a single epoch

            Parameters
            ----------
            batchgen : MultiThreadedAugmenter
                Generator yielding the training batches
            epoch : int
                current epoch

            """

            self.module.train()

            n_batches = batchgen.generator.num_batches * batchgen.num_processes
            pbar = tqdm(enumerate(batchgen), unit=' batch', total=n_batches,
                        desc='Epoch %d' % epoch)

            for batch_nr, batch in pbar:

                data_dict = self._prepare_batch(batch, self.input_device,
                                                self.output_device)

                _, _, _ = self.closure_fn(self.module, data_dict,
                                          optimizers=self.optimizers,
                                          criterions=self.criterions,
                                          metrics=self.metrics,
                                          fold=self.fold,
                                          batch_nr=batch_nr)

            batchgen._finish()

        def predict(self, batchgen, batch_size=None):
            """
            Returns predictions from network for batches from batchgen

            Parameters
            ----------
            batchgen : MultiThreadedAugmenter
                Generator yielding the batches to predict

            batch_size : None or int
                if int: collect batches until batch_size is reached and
                forward them together

            Returns
            -------
            np.ndarray
                predictions from batches
            list of np.ndarray
                labels from batches
            dict
                dictionary containing the mean validation metrics and
                the mean loss values

            """

            self.module.eval()

            outputs_all, labels_all = [], []
            metric_mean_vals = {}
            loss_mean_vals = {}

            n_batches = batchgen.generator.num_batches * batchgen.num_processes

            pbar = tqdm(enumerate(batchgen), unit=' sample',
                        total=n_batches, desc='Test')

            orig_batch_size = batch_size
            batch_list = []

            for i, batch in pbar:

                if not batch_list and (n_batches - i) < batch_size:
                    batch_size = n_batches - i
                    logger.debug("Set Batchsize down to %d to avoid cutting "
                                 "of the last batches" % batch_size)

                data_dict = self._prepare_batch(batch, self.input_device,
                                                self.output_device)
                # queue inputs and labels
                batch_list.append(data_dict)

                # if queue is full process queue:
                if batch_size is None or len(batch_list) >= batch_size:

                    batch_dict = {}
                    for batch in batch_list:
                        for key, val in batch.items():
                            if key in batch_dict.keys():
                                batch_dict[key].append(val)
                            else:
                                batch_dict[key] = [val]

                    for key, val_list in batch_dict.items():
                        batch_dict[key] = torch.cat(val_list)

                    met_vals, loss_vals, preds = self.closure_fn(
                        self.module, batch_dict,
                        optimizers={},
                        criterions=self.criterions,
                        metrics=self.metrics,
                        fold=self.fold)

                    for key, val in met_vals.items():

                        if key in metric_mean_vals.keys():
                            metric_mean_vals[key] += val.detach()
                        else:
                            metric_mean_vals[key] = val.detach()

                    for key, val in loss_vals.items():

                        if key in loss_mean_vals.keys():
                            loss_mean_vals[key] += val.detach()
                        else:
                            loss_mean_vals[key] = val.detach()

                    outputs_all.append(
                        [pytorch_batch_to_numpy(tmp) for tmp in preds])

                    label_dict = {}

                    for key, val in batch_dict.items():
                        if "data" not in key and "img" not in key:
                            label_dict[key] = pytorch_batch_to_numpy(val)

                    labels_all.append([label_dict[key]
                                       for key in sorted(label_dict.keys())])

                    batch_list = []

            batchgen._finish()

            # transpose labels and outputs to have a list of lists of
            # labels of same type
            labels_all = zip(*labels_all)
            outputs_all = zip(*outputs_all)

            labels_all = [np.vstack(_labels) for _labels in labels_all]
            outputs_all = [np.vstack(_outputs) for _outputs in outputs_all]

            # metric_mean_vals contains sums of metrics so far.
            # Dividing by number of batches to get mean values

            # if virtual batchsize is given: calculate actual number of batches
            if batch_size is not None:
                div = np.ceil(n_batches / orig_batch_size)
            else:
                div = n_batches

            val_dict = {}
            for key, val in metric_mean_vals.items():
                val_dict[key] = val / div

            for key, val in loss_mean_vals.items():
                val_dict[key] = val / div

            return outputs_all, labels_all, val_dict

        def save_state(self, file_name, epoch, **kwargs):
            """
            saves the current state via :func:`delira.io.torch.save_checkpoint`

            Parameters
            ----------
            file_name : str
                filename to save the state to
            epoch : int
                current epoch (will be saved for mapping back)
            *args :
                positional arguments
            **kwargs :
                keyword arguments

            """
            save_checkpoint(file_name, self.module, self.optimizers,
                            epoch=epoch, **kwargs)

        @staticmethod
        def load_state(file_name, **kwargs):
            """
            Loads the new state from file via :func:`delira.io.torch.load_checkpoint`

            Parameters
            ----------
            file_name : str
                the file to load the state from
            **kwargs : keyword arguments

            Returns
            -------
            dict
                new state

            """
            return load_checkpoint(file_name, **kwargs)

        def update_state(self, file_name, *args, **kwargs):
            """
            Update internal state from a loaded state

            Parameters
            ----------
            file_name : str
                file containing the new state to load
            *args :
                positional arguments
            **kwargs :
                keyword arguments

            Returns
            -------
            :class:`AbstractNetworkTrainer`
                the trainer with a modified state

            """
            self._update_state(self.load_state(file_name, *args, **kwargs))

        def _update_state(self, new_state):
            """
            Update the state from a given new state

            Parameters
            ----------
            new_state : dict
                new state to update internal state from

            Returns
            -------
            :class:`PyTorchNetworkTrainer`
                the trainer with a modified state

            """
            # print(",".join(new_state.keys()))

            if "model" in new_state:
                self.module.load_state_dict(new_state.pop("model"))

            if "optimizer" in new_state and new_state["optimizer"]:
                optim_state = new_state.pop("optimizer")
                for key in self.optimizers.keys():
                    self.optimizers[key].load_state_dict(
                        optim_state[key])

            if "epoch" in new_state:
                self.start_epoch = new_state.pop("epoch")

            return super()._update_state(new_state)



import abc
import logging
from delira import get_backends

file_logger = logging.getLogger(__name__)


class AbstractNetwork(object):
    """
    Abstract class all networks should be derived from

    """

    _init_kwargs = {}

    @abc.abstractmethod
    def __init__(self, **kwargs):
        """
        Init function to register init kwargs (should be called from all
        subclasses)

        Parameters
        ----------
        **kwargs
            keyword arguments (will be registered to `self.init_kwargs`)

        """
        super().__init__()
        for key, val in kwargs.items():
            self._init_kwargs[key] = val

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        """
        AbstractMethod to specify that each model should be able to be called
        for predictions

        Parameters
        ----------
        *args :
            Positional arguments
        **kwargs :
            Keyword Arguments

        Raises
        ------
        NotImplementedError
            if not overwritten by subclass

        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def closure(model, data_dict: dict, optimizers: dict, criterions={},
                metrics={}, fold=0, **kwargs):
        """
        Function which handles prediction from batch, logging, loss calculation
        and optimizer step
        Parameters
        ----------
        model : :class:`AbstractNetwork`
            model to forward data through
        data_dict : dict
            dictionary containing the data
        optimizers : dict
            dictionary containing all optimizers to perform parameter update
        criterions : dict
            Functions or classes to calculate criterions
        metrics : dict
            Functions or classes to calculate other metrics
        fold : int
            Current Fold in Crossvalidation (default: 0)
        kwargs : dict
            additional keyword arguments

        Returns
        -------
        dict
            Metric values (with same keys as input dict metrics)
        dict
            Loss values (with same keys as input dict criterions)
        list
            Arbitrary number of predictions

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @staticmethod
    def prepare_batch(batch: dict, input_device, output_device):
        """
        Converts a numpy batch of data and labels to suitable datatype and
        pushes them to correct devices

        Parameters
        ----------
        batch : dict
            dictionary containing the batch (must have keys 'data' and 'label'
        input_device :
            device for network inputs
        output_device :
            device for network outputs

        Returns
        -------
        dict
            dictionary containing all necessary data in right format and type
            and on the correct device

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """

        raise NotImplementedError()

    @property
    def init_kwargs(self):
        """
        Returns all arguments registered as init kwargs

        Returns
        -------
        dict
            init kwargs

        """
        return self._init_kwargs

if "TORCH" in get_backends():
    import torch

    class AbstractPyTorchNetwork(AbstractNetwork, torch.nn.Module):
        """
        Abstract Class for PyTorch Networks

        See Also
        --------
        `torch.nn.Module`
        :class:`AbstractNetwork`

        """
        @abc.abstractmethod
        def __init__(self, **kwargs):
            """

            Parameters
            ----------
            **kwargs :
                keyword arguments (are passed to :class:`AbstractNetwork`'s `
                __init__ to register them as init kwargs

            """
            torch.nn.Module.__init__(self)
            AbstractNetwork.__init__(self, **kwargs)

        @abc.abstractmethod
        def forward(self, *inputs):
            """
            Forward inputs through module (defines module behavior)
            Parameters
            ----------
            inputs : list
                inputs of arbitrary type and number

            Returns
            -------
            Any
                result: module results of arbitrary type and number

            """
            raise NotImplementedError()

        def __call__(self, *args, **kwargs):
            """
            Calls Forward method

            Parameters
            ----------
            *args :
                positional arguments (passed to `forward`)
            **kwargs :
                keyword arguments (passed to `forward`)

            Returns
            -------
            Any
                result: module results of arbitrary type and number

            """
            return torch.nn.Module.__call__(self, *args, **kwargs)

        @staticmethod
        def prepare_batch(batch: dict, input_device, output_device):
            """
            Helper Function to prepare Network Inputs and Labels (convert them to
            correct type and shape and push them to correct devices)

            Parameters
            ----------
            batch : dict
                dictionary containing all the data
            input_device : torch.device
                device for network inputs
            output_device : torch.device
                device for network outputs

            Returns
            -------
            dict
                dictionary containing data in correct type and shape and on correct
                device

            """
            return_dict = {"data": torch.from_numpy(batch.pop("data")).to(
                input_device).to(torch.float)}

            for key, vals in batch.items():
                return_dict[key] = torch.from_numpy(vals).to(output_device).to(
                    torch.float)

            return return_dict

if "TF" in get_backends():
    import tensorflow as tf

    class AbstractTfNetwork(AbstractNetwork):
        """
        Abstract Class for Tf Networks

        See Also
        --------
        :class:`AbstractNetwork`

        """

        @abc.abstractmethod
        def __init__(self, sess=tf.Session, **kwargs):
            """

            Parameters
            ----------
            **kwargs :
                keyword arguments (are passed to :class:`AbstractNetwork`'s `
                __init__ to register them as init kwargs

            """
            AbstractNetwork.__init__(self, **kwargs)
            self._sess = sess()
            self.inputs = None
            self.outputs_train = None
            self.outputs_eval = None
            self._losses = None
            self._optims = None
            self.training = True

        def __call__(self, *args):
            """
            Wrapper for calling self.run in eval setting

            Parameters
            ----------
            *args :
                positional arguments (passed to `self.run`)

            Returns
            -------
            Any
                result: module results of arbitrary type and number

            """
            self.training = False
            return self.run(*args)

        def _add_losses(self, losses: dict):
            """
            Add losses to the model graph

            Parameters
            ----------
            losses : dict
                dictionary containing losses.

            """
            raise NotImplementedError()

        def _add_optims(self, optims: dict):
            """
            Add optimizers to the model graph

            Parameters
            ----------
            optims : dict
                dictionary containing losses.
            """
            raise NotImplementedError()

        def run(self, *args):
            """
            Evaluates `self.outputs_train` or `self.outputs_eval` based on `self.training`

            Parameters
            ----------
            *args :
                arguments to feed as `self.inputs`. Must have same length as `self.inputs`

            Returns
            -------
            np.ndarray or list
                based on len(self.outputs*), returns either list or np.ndarray

            """
            if isinstance(self.inputs, tf.Tensor):
                _feed_dict = dict(zip([self.inputs], args))
            else:
                _feed_dict = dict(zip(self.inputs, args))
            if self.training:
                return self._sess.run(self.outputs_train, feed_dict=_feed_dict)
            else:
                return self._sess.run(self.outputs_eval, feed_dict=_feed_dict)



import numpy as np
from batchgenerators.dataloading.data_loader import SlimDataLoaderBase
from multiprocessing import Pool
from .dataset import AbstractDataset
from .sampler import AbstractSampler, SequentialSampler


class BaseDataLoader(SlimDataLoaderBase):
    """
    Class to create a data batch out of data samples

    """
    def __init__(self, dataset: AbstractDataset,
                 batch_size=1, num_batches=None, seed=1,
                 sampler=None):
        """

        Parameters
        ----------
        dataset : AbstractDataset
            dataset to perform sample loading
        batch_size : int
            number of samples per batch
        num_batches : int
            number of batches to load
        seed : int
            seed for Random Number Generator
        sampler : AbstractSampler or None
            class defining the sampling strategy;
            if None: SequentialSampler will be used

        Raises
        ------
        AssertionError
            `sampler` is not :obj:`None` and `sampler` is not an instance of
            the :class:`.sampler.AbstractSampler`

        See Also
        --------
        :class:`.sampler.SequentialSampler`

        """

        # store dataset in self._data
        super().__init__(dataset, batch_size)

        assert isinstance(sampler, AbstractSampler) or sampler is None, \
            "Sampler must be instance of subclass of AbstractSampler of None"

        if sampler is None:
            sampler = SequentialSampler(list(range(len(dataset))))

        self.sampler = sampler

        self.n_samples = len(sampler)
        if num_batches is None:
            num_batches = len(sampler) // batch_size

        self.num_batches = num_batches
        self._seed = seed
        np.random.seed(seed)

        self._batches_generated = 0

    def generate_train_batch(self):
        """
        Generate Indices which behavior based on self.sampling gets data based
        on indices

        Returns
        -------
        dict
            data and labels

        Raises
        ------
        StopIteration
            If the maximum number of batches has been generated
        """

        if self._batches_generated >= self.num_batches:
            raise StopIteration
        else:
            self._batches_generated += 1

            idxs = self.sampler(self.batch_size)

            result = [self._get_sample(_idx) for _idx in idxs]

            result_dict = {}

            # concatenate dict entities by keys
            for _result_dict in result:
                for key, val in _result_dict.items():
                    if key in result_dict.keys():
                        result_dict[key].append(val)
                    else:
                        result_dict[key] = [val]

            # convert list to numpy arrays
            for key, val_list in result_dict.items():
                result_dict[key] = np.asarray(val_list)

            return result_dict

    def _get_sample(self, index):
        """
        Helper functions which returns an element of the dataset

        Parameters
        ----------
        index : int
            index specifying which sample to return

        Returns
        -------
        dict
            Returned Data
        """
        return self._data[index]




from delira import get_backends
import numpy as np
from functools import wraps
import warnings


def dtype_func(class_object):
    """
    Decorator to Check whether the first argument of the decorated function is
    of a certain type

    Parameters
    ----------
    class_object : Any
        type the first function argument should have

    Returns
    -------
    Wrapped Function

    Raises
    ------
    AssertionError
        First argument of decorated function is not of given type

    """
    def instance_checker(func):
        @wraps(func)
        def func_wrapper(checked_object, *args, **kwargs):
            assertion_str = "Argument 1 is not of type %s but of type %s" % \
                            (class_object.__name__,
                             checked_object.__class__.__name__)

            assert isinstance(checked_object, class_object), assertion_str
            return func(checked_object, *args, **kwargs)
        return func_wrapper
    return instance_checker


def classtype_func(class_object):
    """
    Decorator to Check whether the first argument of the decorated function is
    a subclass of a certain type

    Parameters
    ----------
    class_object : Any
        type the first function argument should be subclassed from

    Returns
    -------
    Wrapped Function

    Raises
    ------
    AssertionError
        First argument of decorated function is not a subclass of given type

    """
    def subclass_checker(func):
        @wraps(func)
        def func_wrapper(checked_object, *args, **kwargs):
            assertion_str = "Argument 1 is not subclass of %s but of type %s" \
                            % (class_object.__name__, checked_object.__name__)

            assert issubclass(checked_object, class_object), assertion_str
            return func(checked_object, *args, **kwargs)
        return func_wrapper
    return subclass_checker


def make_deprecated(new_func):
    """
    Decorator which raises a DeprecationWarning for the decorated object

    Parameters
    ----------
    new_func : Any
        new function which should be used instead of the decorated one

    Returns
    -------
    Wrapped Function

    Raises
    ------
    Deprecation Warning

    """
    def deprecation(func):
        @wraps(func)
        def func_wrapper(*args, **kwargs):
            warnings.warn(DeprecationWarning("%s is deprecated in favor of %s"
                                             " and will be removed at next "
                                             "release" % (func.__name__,
                                                          new_func.__name__)))
            return func(*args, **kwargs)

        return func_wrapper
    return deprecation


numpy_array_func = dtype_func(np.ndarray)


if "TORCH" in get_backends():
    import torch
    torch_tensor_func = dtype_func(torch.Tensor)
    torch_module_func = dtype_func(torch.nn.Module)



import os
import logging
import numpy as np
from tqdm.auto import tqdm
import tensorflow as tf
from batchgenerators.dataloading import MultiThreadedAugmenter
from .callbacks import AbstractCallback
from .abstract_trainer import AbstractNetworkTrainer
from .train_utils import create_optims_default_tf as create_optims_default
from .train_utils import initialize_uninitialized
from ..io import tf_load_checkpoint, tf_save_checkpoint
from delira.logging import TrixiHandler
from trixi.logger.tensorboard.tensorboardxlogger import TensorboardXLogger

logger = logging.getLogger(__name__)


class TfNetworkTrainer(AbstractNetworkTrainer):
    """
    Train and Validate a Network

    See Also
    --------
    :class:`AbstractNetwork`

    """

    def __init__(self, network, save_path,
                 losses: dict, optimizer_cls,
                 optimizer_params={}, metrics={}, lr_scheduler_cls=None,
                 lr_scheduler_params={}, gpu_ids=[], save_freq=1,
                 optim_fn=create_optims_default,
                 fold=0, callbacks=[], start_epoch=1,
                 **kwargs):

        """

        Parameters
        ----------
        network : :class:`AbstractTfNetwork`
            the network to train
        save_path : str
            path to save networks to
        losses : dict
            dictionary containing the training losses
        optimizer_cls : subclass of tf.train.Optimizer
            optimizer class implementing the optimization algorithm of choice
        optimizer_params : dict
            keyword arguments passed to optimizer during construction
        metrics : dict
            dictionary containing the validation metrics
        lr_scheduler_cls : Any
            learning rate schedule class: must implement step() method
        lr_scheduler_params : dict
            keyword arguments passed to lr scheduler during construction
        gpu_ids : list
            list containing ids of GPUs to use; if empty: use cpu instead
        save_freq : int
            integer specifying how often to save the current model's state.
            State is saved every state_freq epochs
        optim_fn : function
            creates a dictionary containing all necessary optimizers
        fold : int
            current cross validation fold (0 per default)
        callbacks : list
            initial callbacks to register
        start_epoch : int
            epoch to start training at
        **kwargs :
            additional keyword arguments

        """

        super().__init__(fold, callbacks)

        self.save_path = save_path
        if os.path.isdir(save_path):
            logger.warning(
                "Save Path already exists. Saved Models may be overwritten")
        else:
            os.makedirs(save_path)

        # remove prior Trixihandlers and ensure logging of training results to self.save_path
        # This facilitates visualization of multiple splits/fold inside one tensorboard-instance by means of
        # different tf.Summary.FileWriters()
        root_logger = logging.getLogger()
        for handler in root_logger.handlers:
            handler.close()
        root_logger.handlers = []
        logging.basicConfig(level=logging.INFO,
                            handlers=[TrixiHandler(TensorboardXLogger, 0, self.save_path)])

        self.losses = losses

        self.metrics = metrics

        self.save_freq = save_freq

        # Whether or not to stop the training
        # Used for early stopping
        self.stop_training = False
        self.start_epoch = start_epoch

        self._setup(network, optim_fn, optimizer_cls, optimizer_params,
                    lr_scheduler_cls, lr_scheduler_params, gpu_ids)

        for key, val in kwargs.items():
            setattr(self, key, val)

    def _setup(self, network, optim_fn, optimizer_cls, optimizer_params,
               lr_scheduler_cls, lr_scheduler_params, gpu_ids):
        """
        Defines the Trainers Setup

        Parameters
        ----------
        network : instance of :class: `AbstractTfNetwork`
            the network to train
        optim_fn : function
            creates a dictionary containing all necessary optimizers
        optimizer_cls : subclass of tf.train.Optimizer
            optimizer class implementing the optimization algorithm of choice
        optimizer_params : dict
        lr_scheduler_cls : Any
            learning rate schedule class: must implement step() method
        lr_scheduler_params : dict
            keyword arguments passed to lr scheduler during construction
        gpu_ids : list
            list containing ids of GPUs to use; if empty: use cpu instead
        """

        self.optimizers = optim_fn(optimizer_cls, **optimizer_params)

        # schedulers
        if lr_scheduler_cls is not None:
            for key, optim in self.optimizers.items():
                if not isinstance(lr_scheduler_cls, AbstractCallback):
                    logger.warning("lr_scheduler_cls is not a callback.")
                self.register_callback(lr_scheduler_cls(optim,
                                                        **lr_scheduler_params))

        # asssign closure and prepare batch from network
        self.closure_fn = network.closure

        # TODO: implement multi-GPU and single GPU training with help of
        #  https://www.tensorflow.org/api_docs/python/tf/keras/utils/multi_gpu_model
        #  note: might be bugged in combination with sess.run https://github.com/tensorflow/tensorflow/issues/21788
        #  and https://www.tensorflow.org/api_docs/python/tf/keras/models/clone_model

        """
        if gpu_ids and tf.test.is_gpu_available():
            assert len(gpu_ids) <= len(get_available_gpus()), "more GPUs specified than available"
            self.use_gpu = True
            if len(gpu_ids) > 1:
                logger.warning(
                    "multi-GPU training not yet tested!")

                network.model = tf.keras.utils.multi_gpu_model(
                                        network.model,
                                        len(gpu_ids),
                                        cpu_merge=True,
                                        cpu_relocation=False)
            else:
                network.models = tf.keras.models.clone_model(network.model)
        else:
            self.use_gpu = False
        """
        self.use_gpu = True
        self.module = network

        # TODO: Beautify
        self.module._add_losses(self.losses)
        self.module._add_optims(self.optimizers)
        # check for unitialized variables
        initialize_uninitialized(self.module._sess)

    def train(self, num_epochs, datamgr_train, datamgr_valid=None,
              val_score_key=None, val_score_mode='highest'):
        """
        train network

        Parameters
        ----------
        num_epochs : int
            number of epochs to train
        datamgr_train : BaseDataManager
            Data Manager to create Batch Generator for training
        datamgr_valid : BaseDataManager
            Data Manager to create Batch Generator for validation
        val_score_key : str
            Key of validation metric; must be key in self.metrics
        val_score_mode : str
            String to specify whether a higher or lower validation score is
            optimal; must be in ['highest', 'lowest']

        Returns
        -------
        :class:`AbstractTfNetwork`
            Best model (if `val_score_key` is not a valid key the model of the
            last epoch will be returned)

        """

        self._at_training_begin()

        if val_score_mode == 'highest':
            best_val_score = 0
        elif val_score_mode == 'lowest':
            best_val_score = float('inf')
        else:
            best_val_score = None

        curr_val_score = best_val_score

        self.save_state(os.path.join(self.save_path, "checkpoint_epoch_0"))
        metrics_val = {}

        for epoch in range(self.start_epoch, num_epochs+1):

            self._at_epoch_begin(metrics_val, val_score_key, epoch,
                                 num_epochs)

            batch_gen_train = datamgr_train.get_batchgen(seed=epoch)

            self._train_single_epoch(batch_gen_train, epoch)

            if datamgr_valid:
                # validate with batchsize 1 and 1 augmentation processs to
                # avoid dropping of last elements
                orig_num_aug_processes = datamgr_valid.n_process_augmentation
                orig_batch_size = datamgr_valid.batch_size

                datamgr_valid.batch_size = 1
                datamgr_valid.n_process_augmentation = 1

                labels_val, pred_val, metrics_val = self.predict(
                    datamgr_valid.get_batchgen(), batch_size=orig_batch_size)

                # reset old values
                datamgr_valid.batch_size = orig_batch_size
                datamgr_valid.n_process_augmentation = orig_num_aug_processes

                # ToDO: Move decision, if current model is best to callback
                if val_score_key in metrics_val.keys():
                    curr_val_score = metrics_val[val_score_key]
                    is_best = self._is_better_val_scores(best_val_score,
                                                         curr_val_score,
                                                         val_score_mode)

                else:
                    logger.warning(
                        "Validation score key not in metric dict. "
                        "Logging metrics but can't decide which model is best")

                    is_best = False

                if is_best:
                    best_val_score = curr_val_score
                    tqdm.write(
                        'Best val score = %2.3f' % best_val_score.item())
                else:
                    is_best = False
            else:
                is_best = False
                labels_val, pred_val, metrics_val = {}, {}, {}

            self._at_epoch_end(metrics_val, val_score_key, epoch, is_best)

            # stop training (might be caused by early stopping)
            if self.stop_training:
                break

        return self._at_training_end()

    def _at_training_begin(self, *args, **kwargs):
        """
        Defines behaviour at beginning of training

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        """
        pass

    def _at_training_end(self):
        """
        Defines Behaviour at end of training: Loads best model if available

        Returns
        -------
        :class:`AbstractTfNetwork`
            best network

        """
        if os.path.isfile(os.path.join(self.save_path, 'checkpoint_best.meta')):

            # load best model and return it. Since the state is hidden in the graph, we don't actually need to use
            # self.update_state.
            self.update_state(os.path.join(self.save_path,
                                           'checkpoint_best')
                              )

        return self.module

    def _at_epoch_begin(self, metrics_val, val_score_key, epoch, num_epochs,
                        **kwargs):
        """
        Defines behaviour at beginning of each epoch: Executes all callbacks's
        `at_epoch_begin` method

        Parameters
        ----------
        metrics_val : dict
            validation metrics
        val_score_key : str
            validation score key
        epoch : int
            current epoch
        num_epochs : int
            total number of epochs
        **kwargs :
            keyword arguments

        """

        # execute all callbacks
        for cb in self._callbacks:
            self._update_state(cb.at_epoch_begin(self, val_metrics=metrics_val,
                                                 val_score_key=val_score_key,
                                                 curr_epoch=epoch))

    def _at_epoch_end(self, metrics_val, val_score_key, epoch, is_best,
                      **kwargs):
        """
        Defines behaviour at beginning of each epoch: Executes all callbacks's
        `at_epoch_end` method and saves current state if necessary

        Parameters
        ----------
        metrics_val : dict
            validation metrics
        val_score_key : str
            validation score key
        epoch : int
            current epoch
        num_epochs : int
            total number of epochs
        **kwargs :
            keyword arguments

        """

        for cb in self._callbacks:
            self._update_state(cb.at_epoch_end(self, val_metrics=metrics_val,
                                               val_score_key=val_score_key,
                                               curr_epoch=epoch))

        if epoch % self.save_freq == 0:
            self.save_state(os.path.join(self.save_path,
                                         "checkpoint_epoch_%d" % epoch))

        if is_best:
            self.save_state(os.path.join(self.save_path,
                                         "checkpoint_best"))

    def _train_single_epoch(self, batchgen: MultiThreadedAugmenter, epoch):
        """
        Trains the network a single epoch

        Parameters
        ----------
        batchgen : MultiThreadedAugmenter
            Generator yielding the training batches
        epoch : int
            current epoch

        """
        self.module.training = True

        n_batches = batchgen.generator.num_batches * batchgen.num_processes
        pbar = tqdm(enumerate(batchgen), unit=' batch', total=n_batches,
                    desc='Epoch %d' % epoch)

        for batch_nr, batch in pbar:

            data_dict = batch

            _, _, _ = self.closure_fn(self.module, data_dict,
                                      optimizers=self.optimizers,
                                      losses=self.losses,
                                      metrics=self.metrics,
                                      fold=self.fold,
                                      batch_nr=batch_nr)

        batchgen._finish()

    def predict(self, batchgen, batch_size=None):
        """
        Returns predictions from network for batches from batchgen

        Parameters
        ----------
        batchgen : MultiThreadedAugmenter
            Generator yielding the batches to predict

        batch_size : None or int
            if int: collect batches until batch_size is reached and
            forward them together

        Returns
        -------
        np.ndarray
            predictions from batches
        list of np.ndarray
            labels from batches
        dict
            dictionary containing the mean validation metrics and
            the mean loss values

        """
        self.module.training = False

        outputs_all, labels_all = [], []
        metric_mean_vals = {}
        loss_mean_vals = {}

        n_batches = batchgen.generator.num_batches * batchgen.num_processes

        pbar = tqdm(enumerate(batchgen), unit=' sample',
                    total=n_batches, desc='Test')

        batch_list = []

        orig_batch_size = batch_size

        for i, batch in pbar:

            if not batch_list and (n_batches - i) < batch_size:

                batch_size = n_batches - i
                logger.debug("Set Batchsize down to %d to avoid cutting "
                             "of the last batches" % batch_size)

            # queue inputs and labels
            batch_list.append(batch)

            # if queue is full process queue:
            if batch_size is None or len(batch_list) >= batch_size:

                batch_dict = {}
                for batch in batch_list:
                    for key, val in batch.items():
                        if key in batch_dict.keys():
                            batch_dict[key].append(val)
                        else:
                            batch_dict[key] = [val]

                for key, val_list in batch_dict.items():
                    batch_dict[key] = np.concatenate(val_list)

                met_vals, loss_vals, preds = self.closure_fn(
                    self.module, batch_dict,
                    optimizers={},
                    losses=self.losses,
                    metrics=self.metrics,
                    fold=self.fold)

                for key, val in met_vals.items():

                    if key in metric_mean_vals.keys():
                        metric_mean_vals[key] += val
                    else:
                        metric_mean_vals[key] = val

                for key, val in loss_vals.items():

                    if key in loss_mean_vals.keys():
                        loss_mean_vals[key] += val
                    else:
                        loss_mean_vals[key] = val

                outputs_all.append(tmp for tmp in preds)

                label_dict = {}

                for key, val in batch_dict.items():
                    if "data" not in key and "img" not in key:
                        label_dict[key] = val

                labels_all.append([label_dict[key]
                                   for key in sorted(label_dict.keys())])

                batch_list = []

        batchgen._finish()

        # transpose labels and outputs to have a list of lists of
        # labels of same type
        labels_all = zip(*labels_all)
        outputs_all = zip(*outputs_all)

        labels_all = [np.vstack(_labels) for _labels in labels_all]
        outputs_all = [np.vstack(_outputs) for _outputs in outputs_all]

        # metric_mean_vals contains sums of metrics so far.
        # Dividing by number of batches to get mean values

        # if virtual batchsize is given: calculate actual number of batches
        if batch_size is not None:
            div = np.ceil(n_batches / orig_batch_size)
        else:
            div = n_batches

        val_dict = {}
        for key, val in metric_mean_vals.items():
            val_dict[key] = val / div

        for key, val in loss_mean_vals.items():
            val_dict[key] = val / div

        return outputs_all, labels_all, val_dict

    def save_state(self, file_name):
        """
        saves the current state via :func:`delira.io.tf.save_checkpoint`

        Parameters
        ----------
        file_name : str
            filename to save the state to
        """
        tf_save_checkpoint(file_name, self.module)


    def load_state(self, file_name):
        """
        Loads the new state from file via :func:`delira.io.tf.load_checkpoint`

        Parameters
        ----------
        file_name : str
            the file to load the state from
        Returns
        -------

        """
        return tf_load_checkpoint(file_name, self.module)



import numpy as np
from . import DummyDataset

from delira.data_loading import BaseDataManager
from delira.data_loading import BaseDataLoader, SequentialSampler
from batchgenerators.dataloading import MultiThreadedAugmenter

def test_base_datamanager():

    batch_size = 16

    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    manager = BaseDataManager(dset, batch_size, n_process_augmentation=1, 
                            transforms=None)

    assert isinstance(manager.get_batchgen(), MultiThreadedAugmenter)

    # create batch manually
    data, labels = [], []
    for i in range(batch_size):
        data.append(dset[i]["data"])
        labels.append(dset[i]["label"])

    batch_dict = {"data": np.asarray(data), "label": np.asarray(labels)}

    for key, val in next(manager.get_batchgen()).items():
        assert (val == batch_dict[key]).all()
    for key, val in next(manager.get_batchgen()).items():
        assert len(val) == batch_size

if __name__ == '__main__':
    test_base_datamanager()


__version__ = '0.3.1'

# from .models import AbstractNetwork
# from .logging import TrixiHandler, MultiStreamHandler
# from .data_loading import BaseCacheDataset, BaseLazyDataset, BaseDataManager, \
#     RandomSampler, SequentialSampler
import json
import os
import warnings
warnings.simplefilter('default', DeprecationWarning)
warnings.simplefilter('ignore', ImportWarning)

# to register new pssible backends, they have to be added to this list.
# each backend should consist of a tuple of length 2 with the first entry
# being the package import name and the second being the backend abbreviation.
# E.g. TensorFlow's package is named 'tensorflow' but if the package is found,
# it will be considered as 'tf' later on
__POSSIBLE_BACKENDS = [("torch", "torch"), ("tensorflow", "tf")]
__BACKENDS = []


def _determine_backends():

    _config_file = __file__.replace("__init__.py", ".delira")
    # look for config file to determine backend
    # if file exists: load config into environment variables

    if not os.path.isfile(_config_file):
        _backends = {}
        # try to import all possible backends to determine valid backends

        import importlib
        for curr_backend in __POSSIBLE_BACKENDS:
            try:
                assert len(curr_backend) == 2
                assert all([isinstance(_tmp, str) for _tmp in curr_backend]), \
                    "All entries in current backend must be strings"

                # check if backend can be imported
                bcknd = importlib.util.find_spec(curr_backend[0])

                if bcknd is not None:
                    _backends[curr_backend[1]] = True
                else:
                    _backends[curr_backend[1]] = False
                del bcknd

            except ValueError:
                _backends[curr_backend[1]] = False

        with open(_config_file, "w") as f:
            json.dump({"version": __version__, "backend": _backends},
                      f, sort_keys=True, indent=4)

        del _backends

    # set values from config file to variable
    with open(_config_file) as f:
        _config_dict = json.load(f)
    for key, val in _config_dict.pop("backend").items():
        if val:
            __BACKENDS.append(key.upper())
    del _config_dict

    del _config_file


def get_backends():
    """
    Return List of currently available backends

    """

    if not __BACKENDS:
        _determine_backends()
    return __BACKENDS


# if "TORCH" in get_backends():
#     from .io import torch_load_checkpoint, torch_save_checkpoint
#     from .models import AbstractPyTorchNetwork
#     from .data_loading import TorchvisionClassificationDataset



from delira.logging import MultiStreamHandler

import logging
import numpy as np
import os
import sys

def test_logger():
    handler = MultiStreamHandler(sys.stdout, sys.stderr)

    logging.basicConfig(level=logging.INFO, 
        handlers=[handler])

    logger = logging.getLogger(__name__)
    logger.info("Test Info for Unittest")




import logging
logger = logging.getLogger(__name__)

from delira import get_backends

if "TORCH" in get_backends():
    import torch
    from torchvision import models as t_models

    from delira.models.abstract_network import AbstractPyTorchNetwork


    class GenerativeAdversarialNetworkBasePyTorch(AbstractPyTorchNetwork):
        """Implementation of Vanilla DC-GAN to create 64x64 pixel images

        Notes
        -----
        The fully connected part in the discriminator has been replaced with an
        equivalent convolutional part

        References
        ----------
        https://arxiv.org/abs/1511.06434

        See Also
        --------
        :class:`AbstractPyTorchNetwork`

        """

        def __init__(self, n_channels, noise_length, **kwargs):
            """

            Parameters
            ----------
            n_channels : int
                number of image channels for generated images and input images
            noise_length : int
                length of noise vector
            **kwargs :
                additional keyword arguments

            """

            # register params by passing them as kwargs to parent class __init__
            super().__init__(n_channels=n_channels,
                            noise_length=noise_length,
                            **kwargs)

            gen, discr = self._build_models(n_channels, noise_length, **kwargs)

            self.nz = noise_length

            self.gen = gen
            self.discr = discr

            for key, value in kwargs.items():
                setattr(self, key, value)

        def forward(self, real_image_batch):
            """
            Create fake images by feeding noise through generator and feed results
            and real images through discriminator

            Parameters
            ----------
            real_image_batch : torch.Tensor
                batch of real images

            Returns
            -------
            torch.Tensor
                Generated fake images
            torch.Tensor
                Discriminator prediction of fake images
            torch.Tensor
                Discriminator prediction of real images

            """
            noise = torch.randn(real_image_batch.size(0), self.nz, 1, 1,
                                device=real_image_batch.device)

            fake_image_batch = self.gen(noise)

            discr_pred_fake = self.discr(fake_image_batch)
            discr_pred_real = self.discr(real_image_batch)

            return fake_image_batch, discr_pred_fake, discr_pred_real

        @staticmethod
        def closure(model, data_dict: dict,
                    optimizers: dict, criterions={}, metrics={},
                    fold=0, **kwargs):
            """
            closure method to do a single backpropagation step

            Parameters
            ----------
            model : :class:`ClassificationNetworkBase`
                trainable model
            data_dict : dict
                dictionary containing data
            optimizers : dict
                dictionary of optimizers to optimize model's parameters
            criterions : dict
                dict holding the criterions to calculate errors
                (gradients from different criterions will be accumulated)
            metrics : dict
                dict holding the metrics to calculate
            fold : int
                Current Fold in Crossvalidation (default: 0)
            kwargs : dict
                additional keyword arguments

            Returns
            -------
            dict
                Metric values (with same keys as input dict metrics)
            dict
                Loss values (with same keys as input dict criterions)
            list
                Arbitrary number of predictions as torch.Tensor

            Raises
            ------
            AssertionError
                if optimizers or criterions are empty or the optimizers are not
                specified

            """

            loss_vals = {}
            metric_vals = {}
            total_loss_discr_real = 0
            total_loss_discr_fake = 0
            total_loss_gen = 0

            # choose suitable context manager:
            if optimizers:
                context_man = torch.enable_grad

            else:
                context_man = torch.no_grad

            with context_man():
                batch = data_dict.pop("data")

                # predict batch
                fake_image_batch, discr_pred_fake, discr_pred_real = model(batch)

                # train discr with prediction from real image
                for key, crit_fn in criterions.items():
                    _loss_val = crit_fn(discr_pred_real,
                                        torch.ones_like(discr_pred_real))
                    loss_vals[key + "_discr_real"] = _loss_val.detach()
                    total_loss_discr_real += _loss_val

                # train discr with prediction from fake image
                for key, crit_fn in criterions.items():
                    _loss_val = crit_fn(discr_pred_fake,
                                        torch.zeros_like(discr_pred_fake))
                    loss_vals[key + "_discr_fake"] = _loss_val.detach()
                    total_loss_discr_fake += _loss_val

                total_loss_discr = total_loss_discr_fake + total_loss_discr_real

                if optimizers:

                    # actual backpropagation
                    optimizers["discr"].zero_grad()
                    # perform loss scaling via apex if half precision is enabled
                    with optimizers["discr"].scale_loss(
                            total_loss_discr) as scaled_loss:
                        scaled_loss.backward(retain_graph=True)
                    optimizers["discr"].step()

                # calculate adversarial loss for generator update
                for key, crit_fn in criterions.items():
                    _loss_val = crit_fn(discr_pred_fake,
                                        torch.ones_like(discr_pred_fake))
                    loss_vals[key + "_adversarial"] = _loss_val.detach().cpu()
                    total_loss_gen += _loss_val

                with torch.no_grad():
                    for key, metric_fn in metrics.items():
                        # calculate metrics for discriminator with real prediction
                        metric_vals[key + "_discr_real"] = metric_fn(
                            discr_pred_real,
                            torch.ones_like(
                                discr_pred_real)).detach()

                        # calculate metrics for discriminator with fake prediction
                        metric_vals[key + "_discr_fake"] = metric_fn(
                            discr_pred_fake,
                            torch.zeros_like(
                                discr_pred_fake)).detach()

                        # calculate adversarial metrics
                        metric_vals[key + "_adversarial"] = metric_fn(
                            discr_pred_fake,
                            torch.ones_like(
                                discr_pred_fake)).detach()

                if optimizers:
                    # actual backpropagation
                    optimizers["gen"].zero_grad()
                    # perform loss scaling via apex if half precision is enabled
                    with optimizers["gen"].scale_loss(
                            total_loss_gen) as scaled_loss:
                        scaled_loss.backward()
                    optimizers["gen"].step()

                else:
                    # add prefix "val" in validation mode
                    eval_loss_vals, eval_metrics_vals = {}, {}
                    for key in loss_vals.keys():
                        eval_loss_vals["val_" + str(key)] = loss_vals[key]

                    for key in metric_vals:
                        eval_metrics_vals["val_" + str(key)] = metric_vals[key]

                    loss_vals = eval_loss_vals
                    metric_vals = eval_metrics_vals

            for key, val in {**metric_vals, **loss_vals}.items():
                logging.info({"value": {"value": val.item(), "name": key,
                                        "env_appendix": "_%02d" % fold
                                        }})

            logging.info({'image_grid': {"images": batch, "name": "real_images",
                                        "env_appendix": "_%02d" % fold}})
            logging.info({"image_grid": {"images": fake_image_batch,
                                        "name": "fake_images",
                                        "env_appendix": "_%02d" % fold}})

            return metric_vals, loss_vals, [fake_image_batch, discr_pred_fake,
                                            discr_pred_real]

        @staticmethod
        def _build_models(in_channels, noise_length, **kwargs):
            """
            Builds actual generator and discriminator models

            Parameters
            ----------
            in_channels : int
                number of channels for generated images by generator and inputs of
                discriminator
            noise_length : int
                length of noise vector (generator input)
            **kwargs :
                additional keyword arguments

            Returns
            -------
            torch.nn.Sequential
                generator
            torch.nn.Sequential
                discriminator
            """
            gen = torch.nn.Sequential(
                    # input is Z, going into a convolution
                    torch.nn.ConvTranspose2d(noise_length, 64 * 8, 4, 1, 0, bias=False),
                    torch.nn.BatchNorm2d(64 * 8),
                    torch.nn.ReLU(True),
                    # state size. (64*8) x 4 x 4
                    torch.nn.ConvTranspose2d(64 * 8, 64 * 4, 4, 2, 1, bias=False),
                    torch.nn.BatchNorm2d(64 * 4),
                    torch.nn.ReLU(True),
                    # state size. (64*4) x 8 x 8
                    torch.nn.ConvTranspose2d(64 * 4, 64 * 2, 4, 2, 1, bias=False),
                    torch.nn.BatchNorm2d(64 * 2),
                    torch.nn.ReLU(True),
                    # state size. (64*2) x 16 x 16
                    torch.nn.ConvTranspose2d(64 * 2, 64, 4, 2, 1, bias=False),
                    torch.nn.BatchNorm2d(64),
                    torch.nn.ReLU(True),
                    # state size. (64) x 32 x 32
                    torch.nn.ConvTranspose2d(64, in_channels, 4, 2, 1, bias=False),
                    torch.nn.Tanh()
                    # state size. (nc) x 64 x 64
                )

            discr = torch.nn.Sequential(
                        # input is (nc) x 64 x 64
                        torch.nn.Conv2d(in_channels, 64, 4, 2, 1, bias=False),
                        torch.nn.LeakyReLU(0.2, inplace=True),
                        # state size. (64) x 32 x 32
                        torch.nn.Conv2d(64, 64 * 2, 4, 2, 1, bias=False),
                        torch.nn.BatchNorm2d(64 * 2),
                        torch.nn.LeakyReLU(0.2, inplace=True),
                        # state size. (64*2) x 16 x 16
                        torch.nn.Conv2d(64 * 2, 64 * 4, 4, 2, 1, bias=False),
                        torch.nn.BatchNorm2d(64 * 4),
                        torch.nn.LeakyReLU(0.2, inplace=True),
                        # state size. (64*4) x 8 x 8
                        torch.nn.Conv2d(64 * 4, 64 * 8, 4, 2, 1, bias=False),
                        torch.nn.BatchNorm2d(64 * 8),
                        torch.nn.LeakyReLU(0.2, inplace=True),
                        # state size. (64*8) x 4 x 4
                        torch.nn.Conv2d(64 * 8, 1, 4, 1, 0, bias=False),
                        torch.nn.Sigmoid()
                    )

            return gen, discr



import inspect
import logging
import os
import importlib
from collections import OrderedDict
from itertools import islice
from delira import get_backends

logger = logging.getLogger(__name__)

if "TORCH" in get_backends():

    import torch
    from ..models import AbstractPyTorchNetwork

    def save_checkpoint(file: str, model=None, optimizers={},
                        epoch=None, **kwargs):
        """
        Save model's parameters

        Parameters
        ----------
        file : str
            filepath the model should be saved to
        model : AbstractNetwork or None
            the model which should be saved
            if None: empty dict will be saved as state dict
        optimizers : dict
            dictionary containing all optimizers
        epoch : int
            current epoch (will also be pickled)

        """
        if isinstance(model, torch.nn.DataParallel):
            _model = model.module
        else:
            _model = model

        if isinstance(_model, AbstractPyTorchNetwork):
            model_state = _model.state_dict()
        else:
            model_state = {}
            logger.debug("Saving checkpoint without Model")

        optim_state = OrderedDict()
        for key, val in optimizers.items():
            if isinstance(val, torch.optim.Optimizer):
                optim_state[key] = val.state_dict()

        if not optim_state:
            logger.debug("Saving checkpoint without Optimizer")

        if epoch is None:
            epoch = 0

        state = {"optimizer": optim_state,
                 "model": model_state,
                 "epoch": epoch}

        torch.save(state, file, **kwargs)

    def load_checkpoint(file, **kwargs):
        """
        Loads a saved model

        Parameters
        ----------
        file : str
            filepath to a file containing a saved model
        **kwargs:
            Additional keyword arguments (passed to torch.load)
            Especially "map_location" is important to change the device the
            state_dict should be loaded to

        Returns
        -------
        OrderedDict
            checkpoint state_dict

        """
        checkpoint = torch.load(file, **kwargs)

        if not all([_key in checkpoint
                    for _key in ["model", "optimizer", "epoch"]]):
            return checkpoint['state_dict']
        return checkpoint



from .abstract_sampler import AbstractSampler
from .sequential_sampler import SequentialSampler, \
    PrevalenceSequentialSampler, StoppingPrevalenceSequentialSampler
from .random_sampler import RandomSampler, PrevalenceRandomSampler, \
    StoppingPrevalenceRandomSampler
from .weighted_sampler import WeightedRandomSampler
from .lambda_sampler import LambdaSampler

__all__ = [
    'AbstractSampler',
    'SequentialSampler',
    'PrevalenceSequentialSampler',
    'StoppingPrevalenceSequentialSampler',
    'RandomSampler',
    'PrevalenceRandomSampler',
    'StoppingPrevalenceRandomSampler',
    'WeightedRandomSampler',
    'LambdaSampler'
]


from delira import get_backends

if "TORCH" in get_backends():
    import torch
    import torch.nn.functional as F

    class BCEFocalLossPyTorch(torch.nn.Module):
        """
        Focal loss for binary case without(!) logit

        """

        def __init__(self, alpha=None, gamma=2, reduction='elementwise_mean'):
            """
            Implements Focal Loss for binary class case

            Parameters
            ----------
            alpha : float
                alpha has to be in range [0,1], assigns class weight
            gamma : float
                focusing parameter
            reduction : str
                Specifies the reduction to apply to the output: ‘none’ |
                ‘elementwise_mean’ | ‘sum’. ‘none’: no reduction will be applied,
                ‘elementwise_mean’: the sum of the output will be divided by the
                number of elements in the output, ‘sum’: the output will be summed
            (further information about parameters above can be found in pytorch
            documentation)

            Returns
            -------
            torch.Tensor
                loss value

            """
            super().__init__()
            self.alpha = alpha
            self.gamma = gamma
            self.reduction = reduction

        def forward(self, p, t):
            bce_loss = F.binary_cross_entropy(p, t, reduction='none')

            if self.alpha is not None:
                # create weights for alpha
                alpha_weight = torch.ones(t.shape, device=p.device) * \
                    self.alpha
                alpha_weight = torch.where(torch.eq(t, 1.),
                                           alpha_weight, 1 - alpha_weight)
            else:
                alpha_weight = torch.Tensor([1]).to(p.device)

            # create weights for focal loss
            focal_weight = 1 - torch.where(torch.eq(t, 1.), p, 1 - p)
            focal_weight.pow_(self.gamma)
            focal_weight.to(p.device)

            # compute loss
            focal_loss = focal_weight * alpha_weight * bce_loss

            if self.reduction == 'elementwise_mean':
                return torch.mean(focal_loss)
            if self.reduction == 'none':
                return focal_loss
            if self.reduction == 'sum':
                return torch.sum(focal_loss)
            raise AttributeError('Reduction parameter unknown.')

    class BCEFocalLossLogitPyTorch(torch.nn.Module):
        """
        Focal loss for binary case WITH logit

        """

        def __init__(self, alpha=None, gamma=2, reduction='elementwise_mean'):
            """
            Implements Focal Loss for binary class case

            Parameters
            ----------
            alpha : float
                alpha has to be in range [0,1], assigns class weight
            gamma : float
                focusing parameter
            reduction : str
                Specifies the reduction to apply to the output: ‘none’ |
                ‘elementwise_mean’ | ‘sum’. ‘none’: no reduction will be applied,
                ‘elementwise_mean’: the sum of the output will be divided by the
                number of elements in the output, ‘sum’: the output will be summed
            (further information about parameters above can be found in pytorch
            documentation)

            Returns
            -------
            torch.Tensor
                loss value

            """
            super().__init__()
            self.alpha = alpha
            self.gamma = gamma
            self.reduction = reduction

        def forward(self, p, t):
            bce_loss = F.binary_cross_entropy_with_logits(
                p, t, reduction='none')

            p = torch.sigmoid(p)

            if self.alpha is not None:
                # create weights for alpha
                alpha_weight = torch.ones(t.shape, device=p.device) * \
                    self.alpha
                alpha_weight = torch.where(torch.eq(t, 1.),
                                           alpha_weight, 1 - alpha_weight)
            else:
                alpha_weight = torch.Tensor([1]).to(p.device)

            # create weights for focal loss
            focal_weight = 1 - torch.where(torch.eq(t, 1.), p, 1 - p)
            focal_weight.pow_(self.gamma)
            focal_weight.to(p.device)

            # compute loss
            focal_loss = focal_weight * alpha_weight * bce_loss

            if self.reduction == 'elementwise_mean':
                return torch.mean(focal_loss)
            if self.reduction == 'none':
                return focal_loss
            if self.reduction == 'sum':
                return torch.sum(focal_loss)
            raise AttributeError('Reduction parameter unknown.')



import os


def subdirs(d):
    """For a given directory, return a list of all subdirectories (full paths)

    Parameters
    ----------
    d : string
        given root directory

    Returns
    -------
    list
        list of strings of all subdirectories
    """
    return sorted([os.path.join(d, name) for name in os.listdir(d)
                   if os.path.isdir(os.path.join(d, name))])



import logging
import SimpleITK as sitk
import numpy as np
import json
import os
from abc import abstractmethod

logger = logging.getLogger(__name__)


def load_sample_nii(files, label_load_cls):
    """
    Load sample from multiple ITK files

    Parameters
    ----------
    files : dict with keys `img` and `label`
        filenames of nifti files and label file
    label_load_cls : class
        function to be used for label parsing

    Returns
    -------
    dict
        sample: dict with keys `data` and `label` containing images and label

    Raises
    ------
    AssertionError
        if `img.max()` is greater than 511 or smaller than 1

    """
    img_list = []
    for f in files['img']:
        img = sitk.GetArrayFromImage(sitk.ReadImage(f))
        img = img.astype(np.float32)
        assert img.max() <= 511
        assert img.max() > 1
        img = img/511
        img_list.append(img)
    label_gen = label_load_cls(files['label'])
    label = label_gen.get_labels()
    sample = {"data": np.stack(img_list), "label": label}
    if 'mask' in list(files.keys()):
        mask = sitk.GetArrayFromImage(sitk.ReadImage(files['mask']))
        mask = mask.astype(np.int64)
        sample['mask'] = mask
    return sample


class BaseLabelGenerator(object):
    """
    Base Class to load labels from json files

    """
    def __init__(self, fpath):
        """

        Parameters
        ----------
        fpath : str
            filepath to json file

        Raises
        ------
        AssertionError
            `fpath` does not end with 'json'

        """
        assert fpath.endswith('json')
        self.fpath = fpath
        self.data = self._load()

    def _load(self):
        """
        Private Helper function to load the file

        Returns
        -------
        Any
            loaded values from file

        """
        with open(os.path.join(self.fpath), 'r') as f:
            label = json.load(f)
        return label

    @abstractmethod
    def get_labels(self):
        """
        Abstractmethod to get labels from class

        Raises
        ------
        NotImplementedError
            if not overwritten in subclass

        """
        raise NotImplementedError()



import tensorflow as tf

conv2d = tf.keras.layers.Conv2D
maxpool2d = tf.keras.layers.MaxPool2D
dense = tf.keras.layers.Dense
relu = tf.keras.layers.ReLU
gap2d = tf.keras.layers.GlobalAveragePooling2D
batchnorm2d = tf.keras.layers.BatchNormalization

class ResNet18(tf.keras.Model):
    def __init__(self, num_classes=None):
        super(ResNet18, self).__init__()
        self.conv1 = conv2d(filters=64, strides=2, kernel_size=7,
                            padding='same')
        self.batchnorm1 = batchnorm2d(axis=1)
        self.pool1 = maxpool2d(pool_size=3, strides=2)

        self.conv2_1 = conv2d(filters=64, strides=1, kernel_size=3,
                              padding='same')
        self.conv2_2 = conv2d(filters=64, strides=1, kernel_size=3,
                              padding='same')

        self.conv3_1 = conv2d(filters=128, strides=2, kernel_size=3,
                              padding='same')
        self.conv3_2 = conv2d(filters=128, strides=1, kernel_size=3,
                              padding='same')

        self.conv4_1 = conv2d(filters=256, strides=2, kernel_size=3,
                              padding='same')
        self.conv4_2 = conv2d(filters=256, strides=1, kernel_size=3,
                              padding='same')

        self.conv5_1 = conv2d(filters=512, strides=2, kernel_size=3,
                              padding='same')
        self.conv5_2 = conv2d(filters=512, strides=1, kernel_size=3,
                              padding='same')

        self.gap = gap2d()
        self.dense1 = dense(1000)
        self.dense2 = dense(num_classes)
        self.relu = relu()

    def call(self, inputs, training=None):
        x = self.conv1(inputs)
        x = self.batchnorm1(x)
        x = self.relu(x)
        x = self.pool1(x)

        x = self.conv2_1(x)
        x = self.relu(x)
        x = self.conv2_2(x)
        x = self.relu(x)

        x = self.conv3_1(x)
        x = self.relu(x)
        x = self.conv3_2(x)
        x = self.relu(x)

        x = self.conv4_1(x)
        x = self.relu(x)
        x = self.conv4_2(x)
        x = self.relu(x)

        x = self.conv5_1(x)
        x = self.relu(x)
        x = self.conv5_2(x)
        x = self.relu(x)

        x = self.gap(x)
        x = self.dense1(x)

        x = self.dense2(x)
        return x



from sklearn.metrics import roc_auc_score
from sklearn.metrics import accuracy_score
from ..utils.decorators import make_deprecated

import trixi

from delira import get_backends

if "TORCH" in get_backends():
    import torch

    from .train_utils import pytorch_tensor_to_numpy, float_to_pytorch_tensor 
    @make_deprecated(trixi)
    class AurocMetricPyTorch(torch.nn.Module):
        """
        Metric to Calculate AuROC

        .. deprecated:: 0.1
            :class:`AurocMetricPyTorch` will be removed in next release and is
            deprecated in favor of ``trixi.logging`` Modules

        .. warning::
            :class:`AurocMetricPyTorch` will be removed in next release

        """
        def __init__(self):
            super().__init__()

        def forward(self, outputs: torch.Tensor, targets: torch.Tensor):
            """
            Actual AuROC calculation

            Parameters
            ----------
            outputs : torch.Tensor
                predictions from network
            targets : torch.Tensor
                training targets

            Returns
            -------
            torch.Tensor
                auroc value

            """
            if outputs.dim() == 2:
                outputs = torch.argmax(outputs, dim=1)
            score = roc_auc_score(pytorch_tensor_to_numpy(targets),
                                pytorch_tensor_to_numpy(outputs))
            return float_to_pytorch_tensor(score)


    @make_deprecated(trixi)
    class AccuracyMetricPyTorch(torch.nn.Module):
        """
        Metric to Calculate Accuracy
        
        .. deprecated:: 0.1
            :class:`AccuracyMetricPyTorch` will be removed in next release and is
            deprecated in favor of ``trixi.logging`` Modules

        .. warning::
            class:`AccuracyMetricPyTorch` will be removed in next release

        """
        def __init__(self, normalize=True, sample_weight=None):
            """

            Parameters
            ----------
            normalize : bool, optional (default=True)
            If ``False``, return the number of correctly classified samples.
            Otherwise, return the fraction of correctly classified samples.

            sample_weight : array-like of shape = [n_samples], optional
                Sample weights.

            """
            super().__init__()
            self.normalize = normalize
            self.sample_weight = sample_weight

        def forward(self, outputs: torch.Tensor, targets: torch.Tensor):
            """
            Actual accuracy calcuation

            Parameters
            ----------
            outputs : torch.Tensor
                predictions from network
            targets : torch.Tensor
                training targets

            Returns
            -------
            torch.Tensor
                accuracy value

            """
            outputs = outputs > 0.5
            if outputs.dim() == 2:
                outputs = torch.argmax(outputs, dim=1)
            score = accuracy_score(pytorch_tensor_to_numpy(targets),
                                pytorch_tensor_to_numpy(outputs),
                                self.normalize, self.sample_weight)
            return float_to_pytorch_tensor(score)




from .data_loader import BaseDataLoader
from .data_manager import BaseDataManager
from .dataset import AbstractDataset, BaseCacheDataset, BaseLazyDataset, \
                     ConcatDataset
from .load_utils import default_load_fn_2d
from .sampler import LambdaSampler, \
    WeightedRandomSampler, \
    PrevalenceRandomSampler, \
    RandomSampler, \
    StoppingPrevalenceSequentialSampler, \
    SequentialSampler
from .sampler import __all__ as __all_sampling

from delira import get_backends

if "TORCH" in get_backends():
    from .dataset import TorchvisionClassificationDataset



from .imageops import bounding_box, calculate_origin_offset, max_energy_slice, \
    sitk_new_blank_image, sitk_resample_to_image, sitk_resample_to_shape, \
    sitk_resample_to_spacing

from .path import subdirs
from .time import now

from .config import LookupConfig



import logging
import tensorflow as tf
import typing
from delira.models.abstract_network import AbstractTfNetwork
from delira.models.classification.ResNet18 import ResNet18

tf.keras.backend.set_image_data_format('channels_first')


logger = logging.getLogger(__name__)


class ClassificationNetworkBaseTf(AbstractTfNetwork):
    """
    Implements basic classification with ResNet18

    See Also
    --------
    :class:`AbstractTfNetwork`

    """

    def __init__(self, in_channels: int, n_outputs: int, **kwargs):
        """

        Constructs graph containing model definition and forward pass behavior

        Parameters
        ----------
        in_channels : int
            number of input_channels
        n_outputs : int
            number of outputs (usually same as number of classes)
        """
        # register params by passing them as kwargs to parent class __init__
        super().__init__(in_channels=in_channels,
                         n_outputs=n_outputs,
                         **kwargs)

        # build on CPU for tf.keras.utils.multi_gpu_model() in tf_trainer.py
        # with tf.device('/cpu:0'):
        self.model = self._build_model(n_outputs, **kwargs)

        images = tf.placeholder(shape=[None, in_channels, None, None],
                                dtype=tf.float32)
        labels = tf.placeholder(shape=[None, n_outputs], dtype=tf.float32)

        preds_train = self.model(images, training=True)
        preds_eval = self.model(images, training=False)

        self.inputs = [images, labels]
        self.outputs_train = [preds_train]
        self.outputs_eval = [preds_eval]

        for key, value in kwargs.items():
            setattr(self, key, value)

    def _add_losses(self, losses: dict):
        """
        Adds losses to model that are to be used by optimizers or during evaluation

        Parameters
        ----------
        losses : dict
            dictionary containing all losses. Individual losses are averaged
        """
        if self._losses is not None and len(losses) != 0:
            logging.warning('Change of losses is not yet supported')
            raise NotImplementedError()
        elif self._losses is not None and len(losses) == 0:
            pass
        else:
            self._losses = {}
            for name, _loss in losses.items():
                self._losses[name] = _loss(self.inputs[-1], self.outputs_train[0])

            total_loss = tf.reduce_mean(list(self._losses.values()), axis=0)

            self._losses['total'] = total_loss
            self.outputs_train.append(self._losses)
            self.outputs_eval.append(self._losses)

    def _add_optims(self, optims: dict):
        """
        Adds optims to model that are to be used by optimizers or during training

        Parameters
        ----------
        optim: dict
            dictionary containing all optimizers, optimizers should be of Type[tf.train.Optimizer]
        """
        if self._optims is not None and len(optims) != 0:
            logging.warning('Change of optims is not yet supported')
            pass
            #raise NotImplementedError()
        elif self._optims is not None and len(optims) == 0:
            pass
        else:
            self._optims = optims['default']
            grads = self._optims.compute_gradients(self._losses['total'])
            step = self._optims.apply_gradients(grads)
            self.outputs_train.append(step)

    @staticmethod
    def _build_model(n_outputs: int, **kwargs):
        """
        builds actual model (resnet 18)

        Parameters
        ----------
        n_outputs : int
            number of outputs (usually same as number of classes)
        **kwargs :
            additional keyword arguments
        Returns
        -------
        tf.keras.Model
            created model
        """
        model = ResNet18(num_classes=n_outputs)

        return model

    @staticmethod
    def closure(model: typing.Type[AbstractTfNetwork], data_dict: dict,
                metrics={}, fold=0, **kwargs):
        """
                closure method to do a single prediction.
                This is followed by backpropagation or not based state of
                on model.train

                Parameters
                ----------
                model: AbstractTfNetwork
                    AbstractTfNetwork or its child-clases
                data_dict : dict
                    dictionary containing the data
                metrics : dict
                    dict holding the metrics to calculate
                fold : int
                    Current Fold in Crossvalidation (default: 0)
                **kwargs:
                    additional keyword arguments

                Returns
                -------
                dict
                    Metric values (with same keys as input dict metrics)
                dict
                    Loss values (with same keys as those initially passed to model.init).
                    Additionally, a total_loss key is added
                list
                    Arbitrary number of predictions as np.array

                """

        loss_vals = {}
        metric_vals = {}
        image_names = "input_images"

        inputs = data_dict.pop('data')

        preds, losses, *_ = model.run(inputs, data_dict['label'])

        for key, loss_val in losses.items():
            loss_vals[key] = loss_val

        for key, metric_fn in metrics.items():
            metric_vals[key] = metric_fn(
                preds, *data_dict.values())

        if model.training == False:
            # add prefix "val" in validation mode
            eval_loss_vals, eval_metrics_vals = {}, {}
            for key in loss_vals.keys():
                eval_loss_vals["val_" + str(key)] = loss_vals[key]

            for key in metric_vals:
                eval_metrics_vals["val_" + str(key)] = metric_vals[key]

            loss_vals = eval_loss_vals
            metric_vals = eval_metrics_vals

            image_names = "val_" + str(image_names)

        for key, val in {**metric_vals, **loss_vals}.items():
            logging.info({"value": {"value": val.item(), "name": key,
                                    "env_appendix": "_%02d" % fold
                                    }})

        #logging.info({'image_grid': {"image_array": inputs, "name": image_names,
        #                             "title": "input_images", "env_appendix": "_%02d" % fold}})
        return metric_vals, loss_vals, [preds]



from delira import get_backends

if "TORCH" in get_backends():
    from .torch import save_checkpoint as torch_save_checkpoint
    from .torch import load_checkpoint as torch_load_checkpoint

if "TF" in get_backends():
    from .tf import save_checkpoint as tf_save_checkpoint
    from .tf import load_checkpoint as tf_load_checkpoint



import logging


class MultiStreamHandler(logging.Handler):
    """
    Logging Handler which accepts multiple streams and creates StreamHandlers

    """
    def __init__(self, *streams, level=logging.NOTSET):
        """

        Parameters
        ----------
        streams : streams with write()-attribute
            streams which are passed to StreamHandlers
        level : int (default: logging.NOTSET)
            logging level
        """
        super().__init__(level)

        self._stream_handlers = []

        for _stream in streams:
            self._stream_handlers.append(logging.StreamHandler(_stream))

    def emit(self, record):
        """
        logs the record entity to streams

        Parameters
        ----------
        record : LogRecord
            record to log

        """

        for _handler in self._stream_handlers:
            _handler.emit(record)



import contextlib

from delira import get_backends

if "TORCH" in get_backends():
    import torch

    class DefaultOptimWrapperTorch(object):
        """
        Class wrapping a ``torch`` optimizer to mirror the behavior of ``apex`` 
        without depending on it

        """

        def __init__(self, optimizer: torch.optim.Optimizer, *args, **kwargs):
            """

            Parameters
            ----------
            optimizer : torch.optim.Optimizer
                the actual optimizer to wrap
            *args : 
                additional positional arguments (unused)
            **kwargs : 
                additional keyword arguments (unused)

            """

            self._optimizer = optimizer

        @contextlib.contextmanager
        def scale_loss(self, loss):
            """
            Function which scales the loss in ``apex`` and yields the unscaled loss 
            here to mirror the API

            Parameters
            ----------
            loss : torch.Tensor
                the unscaled loss

            """

            yield loss
            return

        def step(self, closure=None):
            """
            Wraps the step method of the optimizer and calls the original step 
            method

            Parameters
            ----------
            closure : callable
                A closure that reevaluates the model and returns the loss. 
                Optional for most optimizers.

            """

            return self._optimizer.step(closure=closure)

        # Forward any attribute lookups
        def __getattr__(self, attr):
            return getattr(self._optimizer, attr)

        # Forward all torch.optim.Optimizer methods
        def __getstate__(self):
            return self._optimizer.__getstate__()

        def __setstate__(self, *args, **kwargs):
            return self._optimizer.__setstate__(*args, **kwargs)

        def __repr__(self):
            return self._optimizer.__repr__()

        def state_dict(self):
            return self._optimizer.state_dict()

        def load_state_dict(self, state_dict):
            return self._optimizer.load_state_dict(state_dict)

        def zero_grad(self):
            return self._optimizer.zero_grad()

        def add_param_group(self, param_group):
            return self._optimizer.add_param_group(param_group)



from .abstract_callback import AbstractCallback


class EarlyStopping(AbstractCallback):
    """
    Implements Early Stopping as callback

    See Also
    --------
    :class:`AbstractCallback`

    """

    def __init__(self, monitor_key,
                 min_delta=0,
                 patience=0,
                 mode='min'):
        """

        Parameters
        ----------
        monitor_key : str
            the validation key to monitor
        min_delta : float or int
            the minimum difference between the best metric value so far and
            the current one
        patience : int
            number of epochs to wait before stopping training
        mode : str (default: 'min')
            defines the optimum for the monitored value

        """

        super().__init__()

        self.monitor_key = monitor_key,
        self.min_delta = min_delta
        self.patience = patience
        self.mode = mode

        if 'min' == mode:
            self.best_metric = float('inf')
        elif 'max' == mode:
            self.best_metric = - float('inf')

        else:
            raise ValueError("Unknown compare mode: Got %s, but expected one "
                             "of ['min', 'max']" % mode)
        self.epochs_waited = 0

    def _is_better(self, metric):
        """
        Helper function to decide whether the current metric is better than
        the best metric so far

        Parameters
        ----------
        metric :
            current metric value

        Returns
        -------
        bool
            whether this metric is the new best metric or not

        """
        if 'min' == self.mode:
            return metric < (self.best_metric - self.min_delta)
        else:
            return metric > (self.best_metric + self.min_delta)

    def at_epoch_end(self, trainer, **kwargs):
        """
        Actual early stopping: Checks at end of each epoch if monitored metric
        is new best and if it hasn't improved over `self.patience` epochs, the
        training will be stopped

        Parameters
        ----------
        trainer : :class:`AbstractNetworkTrainer`
            the trainer whose arguments can be modified
        **kwargs :
            additional keyword arguments

        Returns
        -------
        :class:`AbstractNetworkTrainer`
            trainer with modified attributes

        """
        metric = kwargs.get("val_metrics", {})[self.monitor_key]

        self.epochs_waited += 1 - int(self._is_better(metric))

        if self.epochs_waited >= self.patience:
            stop_training = True
        else:
            stop_training = False

        return {"stop_training": stop_training}



import logging
file_logger = logging.getLogger(__name__)

from delira import get_backends

if "TORCH" in get_backends():
    import torch
    from torchvision import models as t_models
    from delira.models.abstract_network import AbstractPyTorchNetwork

    class ClassificationNetworkBasePyTorch(AbstractPyTorchNetwork):
        """
        Implements basic classification with ResNet18

        References
        ----------
        https://arxiv.org/abs/1512.03385

        See Also
        --------
        :class:`AbstractPyTorchNetwork`

        """

        def __init__(self, in_channels: int, n_outputs: int, **kwargs):
            """

            Parameters
            ----------
            in_channels : int
                number of input_channels
            n_outputs : int
                number of outputs (usually same as number of classes)

            """
            # register params by passing them as kwargs to parent class __init__
            super().__init__(in_channels=in_channels,
                             n_outputs=n_outputs,
                             **kwargs)

            self.module = self._build_model(in_channels, n_outputs, **kwargs)

            for key, value in kwargs.items():
                setattr(self, key, value)

        def forward(self, input_batch: torch.Tensor):
            """
            Forward input_batch through network

            Parameters
            ----------
            input_batch : torch.Tensor
                batch to forward through network

            Returns
            -------
            torch.Tensor
                Classification Result

            """

            return self.module(input_batch)

        @staticmethod
        def closure(model: AbstractPyTorchNetwork, data_dict: dict,
                    optimizers: dict, criterions={}, metrics={},
                    fold=0, **kwargs):
            """
            closure method to do a single backpropagation step


            Parameters
            ----------
            model : :class:`ClassificationNetworkBasePyTorch`
                trainable model
            data_dict : dict
                dictionary containing the data
            optimizers : dict
                dictionary of optimizers to optimize model's parameters
            criterions : dict
                dict holding the criterions to calculate errors
                (gradients from different criterions will be accumulated)
            metrics : dict
                dict holding the metrics to calculate
            fold : int
                Current Fold in Crossvalidation (default: 0)
            **kwargs:
                additional keyword arguments

            Returns
            -------
            dict
                Metric values (with same keys as input dict metrics)
            dict
                Loss values (with same keys as input dict criterions)
            list
                Arbitrary number of predictions as torch.Tensor

            Raises
            ------
            AssertionError
                if optimizers or criterions are empty or the optimizers are not
                specified

            """

            assert (optimizers and criterions) or not optimizers, \
                "Criterion dict cannot be emtpy, if optimizers are passed"

            loss_vals = {}
            metric_vals = {}
            total_loss = 0

            # choose suitable context manager:
            if optimizers:
                context_man = torch.enable_grad

            else:
                context_man = torch.no_grad

            with context_man():

                inputs = data_dict.pop("data")
                preds = model(inputs)

                if data_dict:

                    for key, crit_fn in criterions.items():
                        _loss_val = crit_fn(preds, *data_dict.values())
                        loss_vals[key] = _loss_val.detach()
                        total_loss += _loss_val

                    with torch.no_grad():
                        for key, metric_fn in metrics.items():
                            metric_vals[key] = metric_fn(
                                preds, *data_dict.values())

            if optimizers:
                optimizers['default'].zero_grad()
                # perform loss scaling via apex if half precision is enabled
                with optimizers["default"].scale_loss(total_loss) as scaled_loss:
                    scaled_loss.backward()
                optimizers['default'].step()

            else:

                # add prefix "val" in validation mode
                eval_loss_vals, eval_metrics_vals = {}, {}
                for key in loss_vals.keys():
                    eval_loss_vals["val_" + str(key)] = loss_vals[key]

                for key in metric_vals:
                    eval_metrics_vals["val_" + str(key)] = metric_vals[key]

                loss_vals = eval_loss_vals
                metric_vals = eval_metrics_vals

            for key, val in {**metric_vals, **loss_vals}.items():
                logging.info({"value": {"value": val.item(), "name": key,
                                        "env_appendix": "_%02d" % fold
                                        }})

            logging.info({'image_grid': {"images": inputs, "name": "input_images",
                                         "env_appendix": "_%02d" % fold}})

            return metric_vals, loss_vals, [preds]

        @staticmethod
        def _build_model(in_channels: int, n_outputs: int, **kwargs):
            """
            builds actual model (resnet 18)

            Parameters
            ----------
            in_channels : int
                number of input channels
            n_outputs : int
                number of outputs (usually same as number of classes)
            **kwargs : dict
                additional keyword arguments

            Returns
            -------
            torch.nn.Module
                created model

            """

            _model = t_models.resnet18(
                pretrained=False, num_classes=n_outputs, **kwargs)
            _model.conv1 = torch.nn.Conv2d(in_channels, 64, kernel_size=7,
                                           stride=2, padding=3, bias=False)

            return _model

        @staticmethod
        def prepare_batch(batch: dict, input_device, output_device):
            """
            Helper Function to prepare Network Inputs and Labels (convert them to
            correct type and shape and push them to correct devices)

            Parameters
            ----------
            batch : dict
                dictionary containing all the data
            input_device : torch.device
                device for network inputs
            output_device : torch.device
                device for network outputs

            Returns
            -------
            dict
                dictionary containing data in correct type and shape and on correct
                device

            """
            return_dict = {"data": torch.from_numpy(batch.pop("data")).to(
                input_device).to(torch.float)}

            for key, vals in batch.items():
                return_dict[key] = torch.from_numpy(vals).to(
                    output_device).squeeze(-1).to(torch.long)

            return return_dict



import abc
import os
from tqdm import tqdm
import numpy as np
from skimage.transform import resize
from sklearn.model_selection import train_test_split
import typing
from ..utils import subdirs
from ..utils.decorators import make_deprecated
from delira import get_backends


class AbstractDataset:
    """
    Base Class for Dataset

    """

    def __init__(self, data_path, load_fn, img_extensions, gt_extensions):
        """

        Parameters
        ----------
        data_path : str
            path to data samples
        load_fn : function
            function to load single sample
        img_extensions : list
            valid extensions of image files
        gt_extensions : list
            valid extensions of label files

        """

        self._img_extensions = img_extensions
        self._gt_extensions = gt_extensions
        self.data_path = data_path
        self._load_fn = load_fn
        self.data = []

    @abc.abstractmethod
    def _make_dataset(self, path):
        """
        Create dataset

        Parameters
        ----------
        path : str
            path to data samples

        Returns
        -------
        list
            data: List of sample paths if lazy; List of samples if not

        """
        pass

    @abc.abstractmethod
    def __getitem__(self, index):
        """
        return data with given index (and loads it before if lazy)

        Parameters
        ----------
        index : int
            index of data

        Returns
        -------
        dict
            data

        """
        pass

    def __len__(self):
        """
        Return number of samples

        Returns
        -------
        int
            number of samples
        """
        return len(self.data)

    def get_sample_from_index(self, index):
        """
        Returns the data sample for a given index
        (without any loading if it would be necessary)
        This implements the base case and can be subclassed
        for index mappings.
        The actual loading behaviour (lazy or cached) should be
        implemented in ``__getitem__``

        See Also
        --------
        :method:ConcatDataset.get_sample_from_index
        :method:BaseLazyDataset.__getitem__
        :method:BaseCacheDataset.__getitem__

        Parameters
        ----------
        index : int
            index corresponding to targeted sample

        Returns
        -------
        Any
            sample corresponding to given index
        """

        return self.data[index]

    def get_subset(self, indices):
        """
        Returns a Subset of the current dataset based on given indices

        Parameters
        ----------
        indices : iterable
            valid indices to extract subset from current dataset

        Returns
        -------
        :class:`BlankDataset`
            the subset

        """

        # extract other important attributes from current dataset
        kwargs = {}

        for key, val in vars(self).items():
            if not (key.startswith("__") and key.endswith("__")):

                if key == "data":
                    continue
                kwargs[key] = val

        kwargs["old_getitem"] = self.__class__.__getitem__
        subset_data = [self.get_sample_from_index(idx) for idx in indices]

        return BlankDataset(subset_data, **kwargs)

    @make_deprecated("Dataset.get_subset")
    def train_test_split(self, *args, **kwargs):
        """
        split dataset into train and test data

        .. deprecated:: 0.3
            method will be removed in next major release

        Parameters
        ----------
        *args :
            positional arguments of ``train_test_split``
        **kwargs :
            keyword arguments of ``train_test_split``

        Returns
        -------
        :class:`BlankDataset`
            train dataset
        :class:`BlankDataset`
            test dataset

        See Also
        --------
        ``sklearn.model_selection.train_test_split``

        """

        train_idxs, test_idxs = train_test_split(
            np.arange(len(self)), *args, **kwargs)

        return self.get_subset(train_idxs), self.get_subset(test_idxs)


class BlankDataset(AbstractDataset):
    """
    Blank Dataset loading the data, which has been passed
    in it's ``__init__`` by it's ``load_fn``

    """

    def __init__(self, data, old_getitem, **kwargs):
        """

        Parameters
        ----------
        data : iterable
            data to load
        old_getitem : function
            get item method of previous dataset
        **kwargs :
            additional keyword arguments (are set as class attribute)

        """
        super().__init__(None, None, None, None)

        self.data = data
        self._old_getitem = old_getitem

        for key, val in kwargs.items():
            setattr(self, key, val)

    def __getitem__(self, index):
        """
        returns single sample corresponding to ``index`` via the ``load_fn``

        Parameters
        ----------
        index : int
            index specifying the data to load

        Returns
        -------
        dict
            dictionary containing a single sample

        """
        return self._old_getitem(self, index)

    def __len__(self):
        """
        returns the length of the dataset

        Returns
        -------
        int
            number of samples

        """

        return len(self.data)


class BaseLazyDataset(AbstractDataset):
    """
    Dataset to load data in a lazy way

    """

    def __init__(self, data_path, load_fn, img_extensions, gt_extensions,
                 **load_kwargs):
        """

        Parameters
        ----------
        data_path : str
            path to data samples
        load_fn : function
            function to load single data sample
        img_extensions : list
            valid extensions of image files
        gt_extensions : list
            valid extensions of label files
        **load_kwargs :
            additional loading keyword arguments (image shape,
            channel number, ...); passed to load_fn

        """
        super().__init__(data_path, load_fn, img_extensions, gt_extensions)
        self._load_kwargs = load_kwargs
        self.data = self._make_dataset(self.data_path)

    def _make_dataset(self, path):
        """
        Helper Function to make a dataset containing paths to all images in a
        certain directory

        Parameters
        ----------
        path : str
            path to data samples

        Returns
        -------
        list
            list of sample paths

        Raises
        ------
        AssertionError
            if `path` is not a valid directory

        """
        data = []
        assert os.path.isdir(path), '%s is not a valid directory' % dir

        for root, _, fnames in sorted(os.walk(path)):
            for fname in fnames:
                fpath = os.path.join(root, fname)
                if self._is_valid_image_file(fpath):
                    sample = [fpath]
                    for ext in self._gt_extensions:
                        gt_path = fpath.rsplit(".", maxsplit=1)[0] + ext
                        if os.path.isfile(gt_path):
                            sample.append(gt_path)

                    data.append(sample)

        return data

    def _is_valid_image_file(self, fname):
        """
        Helper Function to check wheter file is image file and has at least
        one label file

        Parameters
        ----------
        fname : str
            filename of image path

        Returns
        -------
        bool
            is valid data sample

        """
        is_valid_file = False
        for ext in self._img_extensions:
            if fname.endswith(ext):
                is_valid_file = True

        has_label = False
        for ext in self._gt_extensions:
            label_file = fname.rsplit(".", maxsplit=1)[0] + ext
            if os.path.isfile(label_file):
                has_label = True

        return is_valid_file and has_label

    def __getitem__(self, index):
        """
        load data sample specified by index

        Parameters
        ----------
        index : int
            index to specifiy which data sample to load

        Returns
        -------
        dict
            loaded data sample
        """
        data_dict = self._load_fn(*self.get_sample_from_index(index),
                                  **self._load_kwargs)

        return data_dict


class BaseCacheDataset(AbstractDataset):
    """
    Dataset to preload and cache data

    Notes
    -----
    data needs to fit completely into RAM!

    """

    def __init__(self, data_path, load_fn, img_extensions, gt_extensions,
                 **load_kwargs):
        """

        Parameters
        ----------
        data_path : str
            path to data samples
        load_fn : function
            function to load single data sample
        img_extensions : list
            valid extensions of image files
        gt_extensions : list
            valid extensions of label files
        **load_kwargs :
            additional loading keyword arguments (image shape,
            channel number, ...); passed to load_fn

        """
        super().__init__(data_path, load_fn, img_extensions, gt_extensions)
        self._load_kwargs = load_kwargs
        self.data = self._make_dataset(data_path)

    def _make_dataset(self, path):
        """
        Helper Function to make a dataset containing all samples in a certain
        directory

        Parameters
        ----------
        path: str
            path to data samples

        Returns
        -------
        list
            list of sample paths

        Raises
        ------
        AssertionError
            if `path` is not a valid directory

        """
        data = []
        assert os.path.isdir(path), '%s is not a valid directory' % dir

        for root, _, fnames in sorted(os.walk(path)):
            for fname in fnames:
                fpath = os.path.join(root, fname)
                if self._is_valid_image_file(fpath):
                    sample = [fpath]
                    for ext in self._gt_extensions:
                        gt_path = fpath.rsplit(".", maxsplit=1)[0] + ext
                        if os.path.isfile(gt_path):
                            sample.append(gt_path)

                    data.append(self._load_fn(
                        *sample, **self._load_kwargs))

        return data

    def _is_valid_image_file(self, fname):
        """
        Helper Function to check wheter file is image file and has at least
        one label file

        Parameters
        ----------
        fname : str
            filename of image path

        Returns
        -------
        bool
            is valid data sample

        """
        is_valid_file = False
        for ext in self._img_extensions:
            if fname.endswith(ext):
                is_valid_file = True

        has_label = False
        for ext in self._gt_extensions:
            label_file = fname.rsplit(".", maxsplit=1)[0] + ext
            if os.path.isfile(label_file):
                has_label = True

        return is_valid_file and has_label

    def __getitem__(self, index):
        """
        return data sample specified by index

        Parameters
        ----------
        index : int
            index to specifiy which data sample to return

        Returns
        -------
        dict
            data sample

        """
        data_dict = self.get_sample_from_index(index)

        return data_dict


class ConcatDataset(AbstractDataset):
    def __init__(self, *datasets):
        super().__init__(None, None, None, None)

        # check if first item in datasets is list and datasets is of length 1
        if (len(datasets) == 1) and isinstance(datasets[0], list):
            datasets = datasets[0]

        self.data = datasets

    def get_sample_from_index(self, index):
        """
        Returns the data sample for a given index
        (without any loading if it would be necessary)
        This method implements the index mapping of a global index to
        the subindices for each dataset.
        The actual loading behaviour (lazy or cached) should be
        implemented in ``__getitem__``

        See Also
        --------
        :method:AbstractDataset.get_sample_from_index
        :method:BaseLazyDataset.__getitem__
        :method:BaseCacheDataset.__getitem__

        Parameters
        ----------
        index : int
            index corresponding to targeted sample

        Returns
        -------
        Any
            sample corresponding to given index
        """

        curr_max_index = 0
        for dset in self.data:
            prev_max_index = curr_max_index
            curr_max_index += len(dset)

            if prev_max_index <= index < curr_max_index:
                return dset[index - prev_max_index]

            else:
                continue

        raise IndexError("Index %d is out of range for %d items in datasets" %
                         (index, len(self)))

    def __getitem__(self, index):
        return self.get_sample_from_index(index)

    def __len__(self):
        return sum([len(dset) for dset in self.data])


class Nii3DLazyDataset(BaseLazyDataset):
    """
    Dataset to load 3D medical images (e.g. from .nii files) during training

    """

    def __init__(self, data_path, load_fn, img_extensions, gt_extensions,
                 img_files, label_file, **load_kwargs):
        """

        Parameters
        ----------
        data_path : str
            root path to data samples where each samples has it's own folder
        load_fn : function
            function to load single data sample
        img_extensions : list
            valid extensions of image files
        gt_extensions : list
            valid extensions of label files
        img_files : list
            list of image filenames
        label_file : string
            label file name
        **load_kwargs :
            additional loading keyword arguments (image shape,
            channel number, ...); passed to load_fn

        """
        self.img_files = img_files
        self.label_file = label_file
        super().__init__(data_path, load_fn, img_extensions, gt_extensions,
                         **load_kwargs)

    def _make_dataset(self, path):
        """
        Helper Function to make a dataset containing all samples in a certain
        directory

        Parameters
        ----------
        path: str
            path to data samples

        Returns
        -------
        list
            list of sample paths

        Raises
        ------
        AssertionError
            if `path` is not a valid directory

        """
        assert os.path.isdir(path)

        data = [[{'img': [os.path.join(t, i) for i in self.img_files],
                  'label': os.path.join(t, self.label_file)}]
                for t in subdirs(path)]
        return data


class Nii3DCacheDatset(BaseCacheDataset):
    """
    Dataset to load 3D medical images (e.g. from .nii files) before training

    """

    def __init__(self, data_path, load_fn, img_extensions, gt_extensions,
                 img_files, label_file, **load_kwargs):
        """

        Parameters
        ----------
        data_path : str
            root path to data samples where each samples has it's own folder
        load_fn : function
            function to load single data sample
        img_extensions : list
            valid extensions of image files
        gt_extensions : list
            valid extensions of label files
        img_files : list
            list of image filenames
        label_file : str
            label file name
        **load_kwargs :
            additional loading keyword arguments (image shape,
            channel number, ...); passed to load_fn

        """
        self.img_files = img_files
        self.label_file = label_file
        super().__init__(data_path, load_fn, img_extensions, gt_extensions,
                         **load_kwargs)

    def _make_dataset(self, path):
        """
        Helper Function to make a dataset containing all samples in a certain
        directory

        Parameters
        ----------
        path: str
            path to data samples

        Returns
        -------
        list
            list of samples

        Raises
        ------
        AssertionError
            if `path` is not a valid directory

        """
        assert os.path.isdir(path)
        data = []
        for s in tqdm(subdirs(path), unit='samples', desc="Loading samples"):
            files = {'img': [os.path.join(s, i) for i in self.img_files],
                     'label': os.path.join(s, self.label_file)}

            data.append(self._load_fn(files, **self._load_kwargs))
        return data


if "TORCH" in get_backends():
    from torchvision.datasets import CIFAR10, CIFAR100, EMNIST, MNIST, FashionMNIST

    class TorchvisionClassificationDataset(AbstractDataset):
        """
        Wrapper for torchvision classification datasets to provide consistent API

        """

        def __init__(self, dataset, root="/tmp/", train=True, download=True,
                     img_shape=(28, 28), one_hot=False, **kwargs):
            """

            Parameters
            ----------
            dataset : str
                Defines the dataset to use.
                must be one of
                ['mnist', 'emnist', 'fashion_mnist', 'cifar10', 'cifar100']
            root : str
                path dataset (If download is True: dataset will be extracted here;
                else: path to extracted dataset)
            train : bool
                whether to use the train or the testset
            download : bool
                whether or not to download the dataset
                (If already downloaded at specified path,
                it won't be downloaded again)
            img_shape : tuple
                Height and width of output images (will be interpolated)
            **kwargs :
                Additional keyword arguments passed to the torchvision dataset
                class for initialization

            """
            super().__init__("", None, [], [])

            self.download = download
            self.train = train
            self.root = root
            self.img_shape = img_shape
            self.num_classes = None
            self.one_hot = one_hot
            self.data = self._make_dataset(dataset, **kwargs)

        def _make_dataset(self, dataset, **kwargs):
            """
            Create the actual dataset

            Parameters
            ----------
            dataset: str
                Defines the dataset to use.
                must be one of
                ['mnist', 'emnist', 'fashion_mnist', 'cifar10', 'cifar100']
            **kwargs :
                Additional keyword arguments passed to the torchvision dataset
                class for initialization

            Returns
            -------
            torchvision.Dataset
                actual Dataset

            Raises
            ------
            KeyError
                Dataset string does not specify a valid dataset

            """
            if dataset.lower() == "mnist":
                _dataset_cls = MNIST
                self.num_classes = 10
            elif dataset.lower() == "emnist":
                _dataset_cls = EMNIST
                #TODO: EMNIST requires split as kwarg. Search for 'split' in kwargs and
                # update self.num_classes accordingly
                # https://pytorch.org/docs/stable/torchvision/datasets.html#torchvision.datasets.EMNIST
                self.num_classes = None
            elif dataset.lower() == "fashion_mnist":
                _dataset_cls = FashionMNIST
                self.num_classes = 10
            elif dataset.lower() == "cifar10":
                _dataset_cls = CIFAR10
                self.num_classes = 10
            elif dataset.lower() == "cifar100":
                _dataset_cls = CIFAR100
                self.num_classes = 100
            else:
                raise KeyError("Dataset %s not found!" % dataset.lower())

            return _dataset_cls(root=self.root, train=self.train,
                                download=self.download, **kwargs)

        def __getitem__(self, index):
            """
            return data sample specified by index

            Parameters
            ----------
            index : int
                index to specifiy which data sample to return

            Returns
            -------
            dict
                data sample

            """

            data = self.data[index]
            data_dict = {"data": np.array(data[0]),
                         "label": data[1].numpy().reshape(1).astype(np.float32)}

            if self.one_hot:
                #TODO: Remove and refer to batchgenerators transform:
                # https://github.com/MIC-DKFZ/batchgenerators/blob/master/batchgenerators/transforms/utility_transforms.py#L97
                def make_onehot(num_classes, labels):
                    """
                    Function that converts label-encoding to one-hot format.

                    Parameters
                    ----------
                    num_classes : int
                        number of classes present in the dataset
                    
                    labels : np.ndarray
                        labels in label-encoding format
                    
                    Returns
                    -------
                    np.ndarray
                        labels in one-hot format
                    """
                    if isinstance(labels, list) or isinstance(labels, int):
                        labels = np.asarray(labels)
                    assert isinstance(labels, np.ndarray)
                    if len(labels.shape) > 1:
                        one_hot = np.zeros(shape=(list(labels.shape) + [num_classes]),
                                        dtype=labels.dtype)
                        for i, c in enumerate(np.arange(num_classes)):
                            one_hot[..., i][labels == c] = 1
                    else:
                        one_hot = np.zeros(shape=([num_classes]),
                                        dtype=labels.dtype)
                        for i, c in enumerate(np.arange(num_classes)):
                            if labels == c:
                                one_hot[i] = 1
                    return one_hot

                data_dict['label'] = make_onehot(self.num_classes, data_dict['label'])

            img = data_dict["data"]

            img = resize(img, self.img_shape,
                         mode='reflect', anti_aliasing=True)
            if len(img.shape) <= 3:
                img = img.reshape(
                    *img.shape, 1)

            img = img.transpose(
                (len(img.shape) - 1, *range(len(img.shape) - 1)))

            data_dict["data"] = img.astype(np.float32)
            return data_dict

        def __len__(self):
            """
            Return Number of samples

            Returns
            -------
            int
                number of samples

            """
            return len(self.data)



from logging import Handler, NOTSET
from trixi.logger import AbstractLogger as AbstractTrixiLogger


TRIXI_PREFIXES = ["show_", "plot_", "save_", "get_", ""]


class TrixiHandler(Handler):
    """
    Handler to integrate the :mod:`trixi` loggers into the :mod:`logging`
    module

    """

    def __init__(self, logging_cls, level=NOTSET, *args, **kwargs):
        """

        Parameters
        ----------
        logging_cls :
            logging class (must be subclass of `trixi.logger.AbstractLogger`)
        level : int (default: NOTSET)
            logging level
        *args :
            positional arguments to instantiate the logger from the
            `logging_cls`
        **kwargs :
            keyword arguments to instantiate the logger from the `logging_cls`

        """
        super().__init__(level)

        assertion_str = "%s is not a subclass of %s" % (
            logging_cls.__name__, AbstractTrixiLogger.__name__)

        assert issubclass(logging_cls, AbstractTrixiLogger), assertion_str
        self._logger = logging_cls(*args, **kwargs)

    def emit(self, record):
        """
        logs the record entity to `trixi` loggers

        Parameters
        ----------
        record : LogRecord
            record to log

        """
        if not isinstance(record.msg, dict):
            return

        for key, val in record.msg.items():

            for _prefix in TRIXI_PREFIXES:
                if hasattr(self._logger, _prefix + key) and callable(
                        getattr(self._logger, _prefix + key)):

                    if isinstance(val, dict):
                        # get args from val dict
                        args = val.pop("args", [])
                        # combine kwargs from val["kwargs"} and other
                        # key, val pairs in val
                        kwargs = {**val.pop("kwargs", {}),
                                  **val}

                    else:
                        # check if val is iterable
                        try:
                            iter(val)

                            # val is iterable -> use it as args
                            args = val

                        except TypeError:
                            # val is not iterable -> store it in list and use
                            # this list as args
                            args = [val]

                        # val specifies args -> no kwargs given
                        kwargs = {}

                    getattr(self._logger, _prefix + key)(*args, **kwargs)



from scipy.ndimage import zoom
import SimpleITK as sitk
import numpy as np
from .decorators import dtype_func

sitk_img_func = dtype_func(sitk.Image)


def calculate_origin_offset(new_spacing, old_spacing):
    """
    Calculates the origin offset of two spacings

    Parameters
    ----------
    new_spacing : list or np.ndarray or tuple
        new spacing
    old_spacing : list or np.ndarray or tuple
        old spacing

    Returns
    -------
    np.ndarray
        origin offset
    """
    return np.subtract(new_spacing, old_spacing)/2


@sitk_img_func
def sitk_resample_to_spacing(image, new_spacing=(1.0, 1.0, 1.0),
                             interpolator=sitk.sitkLinear,
                             default_value=0.):
    """
    Resamples SITK Image to a given spacing

    Parameters
    ----------
    image : SimpleITK.Image
        image which should be resampled
    new_spacing : list or np.ndarray or tuple
        target spacing
    interpolator : Any
        implements the actual interpolation
    default_value : float
        default value

    Returns
    -------
    SimpleITK.Image
        resampled Image with target spacing

    """
    zoom_factor = np.divide(image.GetSpacing(), new_spacing)
    new_size = np.asarray(np.ceil(np.round(np.multiply(
        zoom_factor, image.GetSize()), decimals=5)), dtype=np.int16)
    offset = calculate_origin_offset(new_spacing, image.GetSpacing())
    reference_image = sitk_new_blank_image(size=new_size,
                                           spacing=new_spacing,
                                           direction=image.GetDirection(),
                                           origin=image.GetOrigin() + offset,
                                           default_value=default_value)
    return sitk_resample_to_image(image, reference_image,
                                  interpolator=interpolator,
                                  default_value=default_value)


@sitk_img_func
def sitk_resample_to_image(image, reference_image, default_value=0.,
                           interpolator=sitk.sitkLinear, transform=None,
                           output_pixel_type=None):
    """
    Resamples Image to reference image

    Parameters
    ----------
    image : SimpleITK.Image
        the image which should be resampled
    reference_image : SimpleITK.Image
        the resampling target
    default_value : float
        default value
    interpolator : Any
        implements the actual interpolation
    transform : Any (default: None)
        transformation
    output_pixel_type : Any (default:None)
        type of output pixels

    Returns
    -------
    SimpleITK.Image
        resampled image

    """
    if transform is None:
        transform = sitk.Transform()
        transform.SetIdentity()
    if output_pixel_type is None:
        output_pixel_type = image.GetPixelID()
    resample_filter = sitk.ResampleImageFilter()
    resample_filter.SetInterpolator(interpolator)
    resample_filter.SetTransform(transform)
    resample_filter.SetOutputPixelType(output_pixel_type)
    resample_filter.SetDefaultPixelValue(default_value)
    resample_filter.SetReferenceImage(reference_image)
    return resample_filter.Execute(image)


def sitk_new_blank_image(size, spacing, direction, origin, default_value=0.):
    """
    Create a new blank image with given properties

    Parameters
    ----------
    size : list or np.ndarray or tuple
        new image size
    spacing : list or np.ndarray or tuple
        spacing of new image
    direction :
        new image's direction
    origin :
        new image's origin
    default_value : float
        new image's default value

    Returns
    -------
    SimpleITK.Image
        Blank image with given properties

    """
    image = sitk.GetImageFromArray(
        np.ones(size, dtype=np.float).T * default_value)
    image.SetSpacing(spacing)
    image.SetDirection(direction)
    image.SetOrigin(origin)
    return image


@sitk_img_func
def sitk_resample_to_shape(img, x, y, z, order=1):
    """
    Resamples Image to given shape

    Parameters
    ----------
    img : SimpleITK.Image
    x : int
        shape in x-direction
    y : int
        shape in y-direction
    z : int
        shape in z-direction
    order : int
        interpolation order

    Returns
    -------
    SimpleITK.Image
        Resampled Image

    """
    img_np = sitk.GetArrayFromImage(img)
    img_np_fixed_size = zoom(img_np,
                             [z / img_np.shape[0],
                              y / img_np.shape[1],
                              x / img_np.shape[2]],
                             order=order)
    return sitk.GetImageFromArray(img_np_fixed_size)


@sitk_img_func
def max_energy_slice(img):
    """Determine the axial slice in which the image energy is max

    Parameters
    ----------
    img : SimpleITK.Image
        given image

    Returns
    -------
    int
        slice index
    """
    assert img.GetDimension() == 3
    return int(np.argmax(np.sum(sitk.GetArrayFromImage(img), axis=(1, 2))))


def sitk_copy_metadata(img_source, img_target):
    """ Copy metadata (=DICOM Tags) from one image to another

    Parameters
    ----------
    img_source : SimpleITK.Image
        Source image
    img_target : SimpleITK.Image
        Source image

    Returns
    -------
    SimpleITK.Image
        Target image with copied metadata
    """
    for k in img_source.GetMetaDataKeys():
        img_target.SetMetaData(k, img_source.GetMetaData(k))
    return img_target


@sitk_img_func
def bounding_box(mask, margin=None):
    """Calculate bounding box coordinates of binary mask

    Parameters
    ----------
    mask : SimpleITK.Image
        Binary mask
    margin : int, default: None
        margin to be added to min/max on each dimension

    Returns
    -------
    tuple
        bounding box coordinates of the form (xmin, xmax, ymin, ymax,
        zmin, zmax)
    """
    # mask_arr is in z, y, x order
    mask_arr = sitk.GetArrayFromImage(mask)
    zdim, ydim, xdim = mask_arr.shape
    nz = np.where(mask_arr != 0)
    lower = [np.min(nz[0]), np.min(nz[1]), np.min(nz[2])]
    upper = [np.max(nz[0]), np.max(nz[1]), np.max(nz[2])]
    if margin is not None:
        for axis in range(3):
            # make sure lower bound with margin is valid
            if lower[axis] - margin >= 0:
                lower[axis] -= margin
            else:
                lower[axis] = 0
            # make sure upper bound with margin is valid
            if upper[axis] + margin <= mask_arr.shape[axis] - 1:
                upper[axis] += margin
            else:
                upper[axis] = mask_arr.shape[axis] - 1
    bbox = lower[0], upper[0], lower[1], upper[1], lower[2], upper[2]
    return bbox



from .abstract_network import AbstractNetwork

from delira import get_backends

if "TORCH" in get_backends():
    from .abstract_network import AbstractPyTorchNetwork
    from .classification import VGG3DClassificationNetworkPyTorch, \
        ClassificationNetworkBasePyTorch

    from .segmentation import UNet2dPyTorch, UNet3dPyTorch

    from .gan import GenerativeAdversarialNetworkBasePyTorch

if "TF" in get_backends():
    from .abstract_network import AbstractTfNetwork
    from .classification import ClassificationNetworkBaseTf


from abc import abstractmethod
from ..dataset import AbstractDataset


class AbstractSampler(object):
    """
    Class to define an abstract Sampling API

    """
    def __init__(self, indices=None):
        pass

    @classmethod
    def from_dataset(cls, dataset: AbstractDataset, **kwargs):
        """
        Classmethod to initialize the sampler from a given dataset

        Parameters
        ----------
        dataset : AbstractDataset
            the given dataset

        Returns
        -------
        :class:`AbstractSampler`
            The initialzed sampler

        """
        indices = list(range(len(dataset)))
        return cls(indices, **kwargs)

    @abstractmethod
    def _get_indices(self, n_indices):
        """
        Function to return a specific number of indices.
        Implements the actual sampling strategy.

        Parameters
        ----------
        n_indices : int
            Number of indices to return

        Returns
        -------
        list
            List with sampled indices

        """
        raise NotImplementedError

    def __call__(self, n_indices):
        """
        Function to call the `get_indices` method of the sampler

        Parameters
        ----------
        n_indices : int
            Number of indices to return

        Returns
        -------
        list
            List with sampled indices

        """
        return self._get_indices(n_indices)

    @abstractmethod
    def __len__(self):
        raise NotImplementedError



import numpy as np

from delira.data_loading import ConcatDataset, BaseCacheDataset

def test_data_subset_concat():
    def load_dummy_sample(path, label_load_fct):
        """
        Returns dummy data, independent of path or label_load_fct
        Parameters
        ----------
        path
        label_load_fct
        Returns
        -------
        : dict
            dict with data and label
        """
        return {'data': np.random.rand(1, 256, 256),
                'label': np.random.randint(2)}

    class DummyCacheDataset(BaseCacheDataset):
        def __init__(self, num: int, label_load_fct, *args, **kwargs):
            """
            Generates random samples with _make_dataset
            Parameters
            ----------
            num : int
                number of random samples
            args :
                passed to BaseCacheDataset
            kwargs :
                passed to BaseCacheDataset

            """
            self.label_load_fct = label_load_fct
            super().__init__(data_path=num, *args, **kwargs)

        def _make_dataset(self, path):
            data = []
            for i in range(path):
                data.append(self._load_fn(i, self.label_load_fct))
            return data

    dset_a = DummyCacheDataset(500, None, load_fn=load_dummy_sample,
                               img_extensions=[], gt_extensions=[])
    dset_b = DummyCacheDataset(700, None, load_fn=load_dummy_sample,
                               img_extensions=[], gt_extensions=[])

    # test concatenating
    concat_dataset = ConcatDataset(dset_a, dset_b)
    assert len(concat_dataset) == (len(dset_a) + len(dset_b))

    assert concat_dataset[0]

    # test slicing:
    half_len_a = len(dset_a) // 2
    half_len_b = len(dset_b) // 2
    assert len(dset_a.get_subset(range(half_len_a))) == half_len_a
    assert len(dset_b.get_subset(range(half_len_b))) == half_len_b

    sliced_concat_set = concat_dataset.get_subset(
        range(half_len_a + half_len_b))
    assert len(sliced_concat_set) == (half_len_a + half_len_b)

    # check if entries are valid
    assert sliced_concat_set[0]


if __name__ == "__main__":
    test_data_subset_concat()



import numpy as np
from ..utils.decorators import dtype_func

from delira import get_backends

if "TORCH" in get_backends():
    import torch

    from ..utils.decorators import torch_tensor_func, torch_module_func

    @torch_tensor_func
    def pytorch_batch_to_numpy(tensor: torch.Tensor):
        """
        Utility Function to cast a whole PyTorch batch to numpy

        Parameters
        ----------
        tensor : torch.Tensor
            the batch to convert

        Returns
        -------
        np.ndarray
            the converted batch

        """
        return np.array([pytorch_tensor_to_numpy(tmp[0]) for tmp in tensor.split(1)])

    @torch_tensor_func
    def pytorch_tensor_to_numpy(tensor: torch.Tensor):
        """
        Utility Function to cast a single PyTorch Tensor to numpy

        Parameters
        ----------
        tensor : torch.Tensor
            the tensor to convert

        Returns
        -------
        np.ndarray
            the converted tensor

        """
        return tensor.detach().cpu().numpy()

    @dtype_func(float)
    def float_to_pytorch_tensor(f: float):
        """
        Converts a single float to a PyTorch Float-Tensor

        Parameters
        ----------
        f : float
            float to convert

        Returns
        -------
        torch.Tensor
            converted float

        """
        return torch.from_numpy(np.array([f], dtype=np.float32))

    @torch_module_func
    def create_optims_default_pytorch(model, optim_cls, **optim_params):
        """
        Function to create a optimizer dictionary
        (in this case only one optimizer for the whole network)

        Parameters
        ----------
        model : :class:`AbstractPyTorchNetwork`
            model whose parameters should be updated by the optimizer
        optim_cls :
            Class implementing an optimization algorithm
        **optim_params :
            Additional keyword arguments (passed to the optimizer class

        Returns
        -------
        dict
            dictionary containing all created optimizers
        """
        return {"default": optim_cls(model.parameters(), **optim_params)}

    @torch_module_func
    def create_optims_gan_default_pytorch(model, optim_cls, **optim_params):
        """
        Function to create a optimizer dictionary
        (in this case two optimizers: One for the generator,
        one for the discriminator)

        Parameters
        ----------
        model : :class:`AbstractPyTorchNetwork`
            model whose parameters should be updated by the optimizer
        optim_cls :
            Class implementing an optimization algorithm
        optim_params :
            Additional keyword arguments (passed to the optimizer class

        Returns
        -------
        dict
            dictionary containing all created optimizers
        """
        return {"gen": optim_cls(model.gen.parameters(), **optim_params),
                "discr": optim_cls(model.discr.parameters(), **optim_params)}

if "TF" in get_backends():
    import tensorflow as tf

    def create_optims_default_tf(optim_cls, **optim_params):
        """
        Function to create a optimizer dictionary
        (in this case only one optimizer)

        Parameters
        ----------
        optim_cls :
            Class implementing an optimization algorithm
        **optim_params :
            Additional keyword arguments (passed to the optimizer class)

        Returns
        -------
        dict
            dictionary containing all created optimizers
        """
        return {"default": optim_cls(**optim_params)}


    def initialize_uninitialized(sess):
        """
        Function to initialize only uninitialized variables in a session graph

        Parameters
        ----------

        sess : tf.Session()
        """

        global_vars = tf.global_variables()
        is_not_initialized = sess.run([tf.is_variable_initialized(var) for var in global_vars])
        not_initialized_vars = [v for (v, f) in zip(global_vars, is_not_initialized) if not f]

        if not_initialized_vars:
            sess.run(tf.variables_initializer(not_initialized_vars))



from abc import abstractmethod
import logging
import pickle

from batchgenerators.dataloading import MultiThreadedAugmenter

from .callbacks import AbstractCallback


logger = logging.getLogger(__name__)

KEYS_TO_GUARD = ["use_gpu",
                 "input_device",
                 "output_device",
                 "_callbacks"]


class AbstractNetworkTrainer(object):
    """
    Defines an abstract API for Network Trainers

    See Also
    --------
    :class:`PyTorchNetworkTrainer`

    """

    def __init__(self, fold=0, callbacks=[]):
        """

        Parameters
        ----------
        fold : int
            current fold in (default: 0)
        callbacks : list
            list of callbacks to register

        """
        self._callbacks = []
        self._fold = fold

        for cbck in callbacks:
            self.register_callback(cbck)

    @abstractmethod
    def _setup(self, *args, **kwargs):
        """
        Defines the actual Trainer Setup

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def _at_training_begin(self, *args, **kwargs):
        """
        Defines the behaviour at beginnig of the training

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def _at_training_end(self, *args, **kwargs):
        """
        Defines the behaviour at the end of the training

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def _at_epoch_begin(self, *args, **kwargs):
        """
        Defines the behaviour at beginnig of each epoch

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def _at_epoch_end(self, *args, **kwargs):
        """
        Defines the behaviour at the end of each epoch

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def _train_single_epoch(self, batchgen: MultiThreadedAugmenter, epoch):
        """
        Defines a routine to train a single epoch

        Parameters
        ----------
        batchgen : MultiThreadedAugmenter
            generator holding the batches
        epoch : int
            current epoch

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def train(self, num_epochs, datamgr_train, datamgr_valid=None,
              val_score_key=None, val_score_mode='highest'):
        """
        Defines a routine to train a specified number of epochs

        Parameters
        ----------
        num_epochs : int
            number of epochs to train
        datamgr_train : DataManager
            the datamanager holding the train data
        datamgr_valid : DataManager
            the datamanager holding the validation data (default: None)
        val_score_key : str
            the key specifying which metric to use for validation
            (default: None)
        val_score_mode : str
            key specifying what kind of validation score is best

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def predict(self, batchgen, batchsize=None):
        """
        Defines a rotine to predict data obtained from a batchgenerator

        Parameters
        ----------
        batchgen : MultiThreadedAugmenter
            Generator Holding the Batches
        batchsize : Artificial batchsize (sampling will be done with batchsize
                    1 and sampled data will be stacked to match the artificial
                    batchsize)(default: None)

        Raises
        ------
        NotImplementedError
            If not overwritten by subclass

        """
        raise NotImplementedError()

    @property
    def fold(self):
        """
        Get current fold

        Returns
        -------
        int
            current fold

        """
        return self._fold

    @fold.setter
    def fold(self, fold):
        """
        Set the current fold

        Parameters
        ----------
        fold : int
            new fold

        Raises
        ------
        ValueError
            if `fold` is not covertable to :obj:`int`

        """
        try:
            self._fold = int(fold)

        except ValueError as e:
            logger.error(e)
            raise e

    def __setattr__(self, key, value):
        """
        Set attributes and guard specific attributes after they have been set
        once

        Parameters
        ----------
        key : str
            the attributes name
        value : Any
            the value to set

        Raises
        ------
        PermissionError
            If attribute which should be set is guarded

        """

        # check if key has been set once
        if key in KEYS_TO_GUARD and hasattr(self, key):
            raise PermissionError("%s should not be overwritten after "
                                  "it has been set once" % key)
        else:
            super().__setattr__(key, value)

    def register_callback(self, callback: AbstractCallback):
        """
        Register Callback to Trainer

        Parameters
        ----------
        callback : :class:`AbstractCallback`
            the callback to register

        Raises
        ------
        AssertionError
            `callback` is not an instance of :class:`AbstractCallback` and has
            not both methods ['at_epoch_begin', 'at_epoch_end']

        """
        assertion_str = "Given callback is not valid; Must be instance of " \
                        "AbstractCallback or provide functions " \
                        "'at_epoch_begin' and 'at_epoch_end'"

        assert isinstance(callback, AbstractCallback) or \
            (hasattr(callback, "at_epoch_begin")
             and hasattr(callback, "at_epoch_end")), assertion_str

        self._callbacks.append(callback)

    def save_state(self, file_name, *args, **kwargs):
        """
        saves the current state

        Parameters
        ----------
        file_name : str
            filename to save the state to
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        """
        with open(file_name, "wb") as f:
            pickle.dump(vars(self), f, *args, **kwargs)

    @staticmethod
    def load_state(file_name, *args, **kwargs):
        """
        Loads the new state from file

        Parameters
        ----------
        file_name : str
            the file to load the state from
        *args :
            positional arguments
        **kwargs : keyword arguments

        Returns
        -------
        dict
            new state

        """
        with open(file_name, "rb") as f:
            new_state = pickle.load(f, *args, **kwargs)

        return new_state

    def _update_state(self, new_state):
        """
        Update the state from a given new state

        Parameters
        ----------
        new_state : dict
            new state to update internal state from

        Returns
        -------
        :class:`AbstractNetworkTrainer`
            the trainer with a modified state

        """
        for key, val in new_state.items():
            if (key.startswith("__") and key.endswith("__")):
                continue

            try:
                setattr(self, key, val)

            except PermissionError:
                logger.error("Trying to overwrite attribute %s of "
                             "NetworkTrainer, which is not allowed!" % key)

        return self

    def update_state(self, file_name, *args, **kwargs):
        """
        Update internal state from a loaded state

        Parameters
        ----------
        file_name : str
            file containing the new state to load
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Returns
        -------
        :class:`AbstractNetworkTrainer`
            the trainer with a modified state

        """
        self._update_state(self.load_state(file_name, *args, **kwargs))

    @staticmethod
    def _is_better_val_scores(old_val_score, new_val_score,
                              mode='highest'):
        """
        Check whether the new val score is better than the old one
        with respect to the optimization goal

        Parameters
        ----------
        old_val_score :
            old validation score
        new_val_score :
            new validation score
        mode: str
            String to specify whether a higher or lower validation score is
            optimal; must be in ['highest', 'lowest']

        Returns
        -------
        bool
            True if new score is better, False otherwise
        """

        assert mode in ['highest', 'lowest'], "Invalid Comparison Mode"

        if mode == 'highest':
            return new_val_score > old_val_score
        elif mode == 'lowest':
            return new_val_score < old_val_score



import datetime


def now():
    """Return current time as YYYY-MM-DD_HH-MM-SS

    Returns
    -------
    string
        current time
    """

    return datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')



from .abstract_callback import AbstractCallback
from .early_stopping import EarlyStopping

from delira import get_backends

if "TORCH" in get_backends():
    from .pytorch_schedulers import DefaultPyTorchSchedulerCallback
    from .pytorch_schedulers import CosineAnnealingLRCallback as \
        CosineAnnealingLRCallbackPyTorch
    from .pytorch_schedulers import ExponentialLRCallback as \
        ExponentialLRCallbackPyTorch

    from .pytorch_schedulers import LambdaLRCallback as LambdaLRCallbackPyTorch
    from .pytorch_schedulers import MultiStepLRCallback as \
        MultiStepLRCallbackPyTorch
    from .pytorch_schedulers import ReduceLROnPlateauCallback as \
        ReduceLROnPlateauCallbackPyTorch
    from .pytorch_schedulers import StepLRCallback as StepLRCallbackPyTorch



from delira import get_backends
import numpy as np
import pytest
import time
import gc
import sys
from psutil import virtual_memory

if "TORCH" in get_backends():
    from delira.models import AbstractPyTorchNetwork, UNet2dPyTorch, \
        UNet3dPyTorch, ClassificationNetworkBasePyTorch, \
        VGG3DClassificationNetworkPyTorch, GenerativeAdversarialNetworkBasePyTorch
    from delira.training.train_utils import create_optims_default_pytorch, \
        create_optims_gan_default_pytorch
    from delira.utils.context_managers import DefaultOptimWrapperTorch
    import torch

    test_cases = [
        # UNet 2D
        (
            UNet2dPyTorch(5, in_channels=1),  # model
            (1, 32, 32),  # input shape
            (32, 32),  # output shape
            {"loss_fn": torch.nn.CrossEntropyLoss()
             },  # loss function
            create_optims_default_pytorch,  # optim_fn
            4,  # output range (num_classes -1)
            True  # half precision
        ),
        # UNet 3D
        (
            UNet3dPyTorch(5, in_channels=1),  # model
            (1, 32, 32, 32),  # input shape
            (32, 32, 32),  # output shape
            {"loss_fn": torch.nn.CrossEntropyLoss()
             },  # loss function
            create_optims_default_pytorch,  # optim_fn
            4,  # output range (num_classes) - 1
            True  # half precision
        ),
        # Base Classifier (Resnet 18)
        (
            ClassificationNetworkBasePyTorch(1, 10),
            # model
            (1, 224, 224),  # input shape
            9,  # output shape (num_classes - 1)
            {"loss_fn": torch.nn.CrossEntropyLoss()
             },  # loss function
            create_optims_default_pytorch,  # optim_fn
            None,  # no max_range needed
            True  # half precision
        ),
        # 3D VGG
        (
            VGG3DClassificationNetworkPyTorch(1, 10),
            # model
            (1, 32, 224, 224),  # input shape
            9,  # output shape (num_classes - 1)
            {"loss_fn": torch.nn.CrossEntropyLoss()
             },  # loss function
            create_optims_default_pytorch,  # optim fn
            None,  # no max_range needed
            True  # half precision
        ),
        # DCGAN
        (
            GenerativeAdversarialNetworkBasePyTorch(
                1, 100),
            # model
            (1, 64, 64),  # input shape
            (1, 1),  # arbitrary shape (not needed)
            {"loss_fn": torch.nn.MSELoss()},  # loss
            create_optims_gan_default_pytorch,
            # optimizer function
            1,  # standard max range
            True  # half precision
        )
    ]
else:
    # tests will be skipped anyway, arguments don't matter
    test_cases = [[None] * 7]


@pytest.mark.parametrize("model,input_shape,target_shape,loss_fn,"
                         "create_optim_fn,max_range,half_precision",
                         test_cases)
@pytest.mark.skipif((virtual_memory().total / 1024.**3) < 20,
                    reason="Less than 20GB of memory")
@pytest.mark.skipif("TORCH" not in get_backends(),
                    reason="torch backend not installed")
def test_pytorch_model_default(model, input_shape,
                               target_shape, loss_fn, create_optim_fn,
                               max_range, half_precision):

    try:
        from apex import amp
        amp_handle = amp.init(half_precision)
        wrapper_fn = amp_handle.wrap_optimizer
    except ImportError:
        wrapper_fn = DefaultOptimWrapperTorch

    start_time = time.time()

    # test backward if optimizer fn is not None
    if create_optim_fn is not None:
        optim = {k: wrapper_fn(v, num_loss=len(loss_fn))
                 for k, v in create_optim_fn(model, torch.optim.Adam).items()}

    else:
        optim = {}

    closure = model.closure
    device = torch.device("cpu")
    model = model.to(device)
    prepare_batch = model.prepare_batch

    # classification label: target_shape specifies max label
    if isinstance(target_shape, int):
        label = np.asarray([np.random.randint(target_shape) for i in range(
            10)])
    else:
        label = np.random.rand(10, *target_shape) * max_range

    data_dict = {
        "data": np.random.rand(10, *input_shape),
        "label": label
    }

    try:
        data_dict = prepare_batch(data_dict, device, device)
        closure(model, data_dict, optim, loss_fn, {})
    except Exception as e:
        assert False, "Test for %s not passed: Error: %s" \
                      % (model.__class__.__name__, e)

    end_time = time.time()

    print("Time needed for %s: %.3f" % (model.__class__.__name__, end_time -
                                        start_time))

    del device
    del optim
    del closure
    del prepare_batch
    del model
    try:
        del amp_handle
    except NameError:
        pass
    gc.collect()


if __name__ == '__main__':
    # checks if networks are valid (not if they learn something)
    test_pytorch_model_default()



from delira import get_backends
import pytest

@pytest.mark.skipif("TF" not in get_backends(),
                    reason="No TF Backend Installed")
def test_load_save():
    from delira.io import tf_load_checkpoint, tf_save_checkpoint
    from delira.models import AbstractTfNetwork
    from delira.training.train_utils import initialize_uninitialized
    import tensorflow as tf
    import numpy as np

    class DummyNetwork(AbstractTfNetwork):
        def __init__(self, in_channels, n_outputs):
            super().__init__(in_channels=in_channels, n_outputs=n_outputs)
            self.net = self._build_model(in_channels, n_outputs)

        @staticmethod
        def _build_model(in_channels, n_outputs):
            return tf.keras.models.Sequential(
                layers=[
                    tf.keras.layers.Dense(
                        64, input_shape=in_channels, bias_initializer='glorot_uniform'),
                    tf.keras.layers.ReLU(),
                    tf.keras.layers.Dense(n_outputs, bias_initializer='glorot_uniform')]
            )

    net = DummyNetwork((32,), 1)
    initialize_uninitialized(net._sess)

    vars_1 = net._sess.run(tf.global_variables())

    tf_save_checkpoint("./model", model=net)

    net._sess.run(tf.initializers.global_variables())

    vars_2 = net._sess.run(tf.global_variables())

    tf_load_checkpoint("./model", model=net)

    vars_3 = net._sess.run(tf.global_variables())

    for var_1, var_2 in zip(vars_1, vars_2):
        assert np.all(var_1 != var_2)

    for var_1, var_3 in zip(vars_1, vars_3):
        assert np.all(var_1 == var_3)

    if __name__ == '__main__':
        test_load_save()



from delira import get_backends
import numpy as np
import pytest
import time
import gc
import sys
from psutil import virtual_memory

if "TF" in get_backends():
    from delira.models import ClassificationNetworkBaseTf
    from delira.training.train_utils import create_optims_default_tf
    import tensorflow as tf

    test_cases = [(
        ClassificationNetworkBaseTf(1, 10),
        # model
        (1, 224, 224),  # input shape
        9,  # output shape (num_classes - 1)
        {"loss_fn": tf.losses.softmax_cross_entropy
         },  # loss function
        create_optims_default_tf,  # optim_fn
        4
    )]
else:
    # tests will be skipped anyways, arguments don't matter
    test_cases = [[None] * 6]


@pytest.mark.parametrize("model,input_shape,target_shape,loss_fn,"
                         "create_optim_fn,max_range",
                         test_cases)
@pytest.mark.skipif((virtual_memory().total / 1024.**3) < 20,
                    reason="Less than 20GB of memory")
@pytest.mark.skipif("TF" not in get_backends(),
                    reason="No TF Backend Installed")
def test_tf_model_default(model, input_shape,
                          target_shape, loss_fn, create_optim_fn, max_range):

    start_time = time.time()

    # test backward if optimizer fn is not None
    if create_optim_fn is not None:
        optim = create_optim_fn(tf.train.AdamOptimizer)

    else:
        optim = {}

    model._add_losses(loss_fn)
    model._add_optims(optim)
    model._sess.run(tf.initializers.global_variables())

    closure = model.closure

    # classification label: target_shape specifies max label
    if isinstance(target_shape, int):
        label = np.asarray([np.random.randint(target_shape) for i in range(
            10)])[np.newaxis, :]
    else:
        label = np.random.rand(10, *target_shape) * max_range

    data_dict = {
        "data": np.random.rand(10, *input_shape),
        "label": label
    }

    try:
        closure(model, data_dict)
    except Exception as e:
        assert False, "Test for %s not passed: Error: %s" \
            % (model.__class__.__name__, e)

    end_time = time.time()

    print("Time needed for %s: %.3f" % (model.__class__.__name__, end_time -
                                        start_time))

    del optim
    del closure
    del model
    gc.collect()

    if __name__ == '__main__':
        # checks if networks are valid (not if they learn something)
        test_tf_model_default()



from delira.logging import TrixiHandler

from trixi.logger import NumpyPlotFileLogger
import logging
import numpy as np
import os

def test_trixi_logger():
    handler = TrixiHandler(NumpyPlotFileLogger, img_dir="./imgs", plot_dir="./plots")

    logging.basicConfig(level=logging.INFO, 
        handlers=[handler])

    logger = logging.getLogger(__name__)
    logger.info({'image': {"image": np.random.rand(28, 28), "name": "test_img"}})


if __name__ == '__main__':
    test_trixi_logger()


from random import choices
from ..dataset import AbstractDataset
from .abstract_sampler import AbstractSampler


class WeightedRandomSampler(AbstractSampler):
    """
    Implements Weighted Random Sampling

    """
    def __init__(self, indices, weights=None, cum_weights=None):
        """

        Parameters
        ----------
        indices : list
            list of classes each sample belongs to. List index corresponds to
            data index and the value at a certain index indicates the
             corresponding class
        weights : Any or None
            sampling weights; for more details see random.choices
        cum_weights : Any or None
            cumulative sampling weights; for more details see random.choices

        """
        super().__init__()

        self._indices = list(range(len(indices)))
        self._weights = weights
        self._cum_weights = cum_weights
        self._global_index = 0

    @classmethod
    def from_dataset(cls, dataset: AbstractDataset, **kwargs):
        """

        Classmethod to initialize the sampler from a given dataset

        Parameters
        ----------
        dataset : AbstractDataset
            the given dataset

        Returns
        -------
        AbstractSampler
            The initialzed sampler

        """

        indices = list(range(len(dataset)))
        labels = [d['label'] for d in dataset.data]
        return cls(labels, **kwargs)

    def _get_indices(self, n_indices):
        """
        Actual Sampling

        Parameters
        ----------
        n_indices : int
            number of indices to return

        Returns
        -------
        list
            list of sampled indices

        Raises
        ------
        StopIteration
            If maximal number of samples is reached
        TypeError
            if weights and cum_weights are specified at the same time
        ValueError
            if weights or cum_weights don't match the population

        """
        if self._global_index >= len(self._indices):
            self._global_index = 0
            raise StopIteration

        new_global_idx = self._global_index + n_indices

        # If we reach end, make batch smaller
        if new_global_idx >= len(self._indices):
            new_global_idx = len(self._indices)

        samples = choices(self._indices, weights=self._weights,
                          cum_weights=self._cum_weights,
                          k=new_global_idx-self._global_index)

        self._global_index = new_global_idx
        return samples

    def __len__(self):
        return len(self._indices)




from delira.data_loading import BaseDataLoader, SequentialSampler

import numpy as np 
from . import DummyDataset

def test_data_loader():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])
    sampler = SequentialSampler.from_dataset(dset)
    loader = BaseDataLoader(dset, batch_size=16, sampler=sampler)

    assert isinstance(loader.generate_train_batch(), dict)
    for key, val in loader.generate_train_batch().items():
        assert len(val) == 16
    assert "label" in loader.generate_train_batch()
    assert "data" in loader.generate_train_batch()
    assert len(set([_tmp for _tmp in loader.generate_train_batch()["label"]])) == 1

if __name__ == '__main__':
    test_data_loader()